"""LLM predictor tests."""
from typing import Any
from unittest.mock import patch

from llama_index.llm_predictor.structured import LLMPredictor, StructuredLLMPredictor
from llama_index.prompts import BasePromptTemplate
from llama_index.prompts.base import PromptTemplate
from llama_index.types import BaseOutputParser

try:
    gptcache_installed = True
except ImportError:
    gptcache_installed = False


class MockOutputParser(BaseOutputParser):
    """Mock output parser."""

    def parse(self, output: str) -> str:
        """Parse output."""
        return output + "\n" + output

    def format(self, output: str) -> str:
        """Format output."""
        return output


def mock_llmpredictor_predict(prompt: BasePromptTemplate, **prompt_args: Any) -> str:
    """Mock LLMPredictor predict."""
    return prompt_args["query_str"]


@patch.object(LLMPredictor, "predict", side_effect=mock_llmpredictor_predict)
@patch.object(LLMPredictor, "__init__", return_value=None)
def test_struct_llm_predictor(mock_init: Any, mock_predict: Any) -> None:
    """Test LLM predictor."""
    llm_predictor = StructuredLLMPredictor()
    output_parser = MockOutputParser()
    prompt = PromptTemplate("{query_str}", output_parser=output_parser)
    llm_prediction = llm_predictor.predict(prompt, query_str="hello world")
    assert llm_prediction == "hello world\nhello world"

    # no change
    prompt = PromptTemplate("{query_str}")
    llm_prediction = llm_predictor.predict(prompt, query_str="hello world")
    assert llm_prediction == "hello world"


# TODO: bring back gptcache tests
# @pytest.mark.skipif(not gptcache_installed, reason="gptcache not installed")
# def test_struct_llm_predictor_with_cache() -> None:
#     """Test LLM predictor."""
#     from gptcache.processor.pre import get_prompt
#     from gptcache.manager.factory import get_data_manager
#     from llama_index.bridge.langchain import GPTCache

#     def init_gptcache_map(cache_obj: Cache) -> None:
#         cache_path = "test"
#         if os.path.isfile(cache_path):
#             os.remove(cache_path)
#         cache_obj.init(
#             pre_embedding_func=get_prompt,
#             data_manager=get_data_manager(data_path=cache_path),
#         )

#     responses = ["helloworld", "helloworld2"]

#     llm = FakeListLLM(responses=responses)
#     predictor = LLMPredictor(llm, False, GPTCache(init_gptcache_map))

#     prompt = DEFAULT_SIMPLE_INPUT_PROMPT
#     llm_prediction = predictor.predict(prompt, query_str="hello world")
#     assert llm_prediction == "helloworld"

#     # due to cached result, faked llm is called only once
#     llm_prediction = predictor.predict(prompt, query_str="hello world")
#     assert llm_prediction == "helloworld"

#     # no cache, return sequence
#     llm.cache = False
#     llm_prediction = predictor.predict(prompt, query_str="hello world")
#     assert llm_prediction == "helloworld2"



"""LLM Prompt Program."""
from abc import abstractmethod
from typing import Any, Generic, Optional, Type, TypeVar

from llama_index.bridge.pydantic import BaseModel

from llama_index.program.base_program import BasePydanticProgram
from llama_index.prompts.base import PromptTemplate
from llama_index.types import Model

LM = TypeVar("LM")


class BaseLLMFunctionProgram(BasePydanticProgram[BaseModel], Generic[LM]):
    """Base LLM Prompt Program.

    This is a base class for LLM endpoints that can return
    a structured output given the prompt.

    NOTE: this only works for structured endpoints atm
    (does not work for text completion endpoints.)

    """

    @classmethod
    @abstractmethod
    def from_defaults(
        cls,
        output_cls: Type[Model],
        prompt_template_str: Optional[str] = None,
        prompt: Optional[PromptTemplate] = None,
        llm: Optional[LM] = None,
        **kwargs: Any,
    ) -> "BaseLLMFunctionProgram":
        """Initialize program from defaults."""



import langchain

# prompts
from langchain import BasePromptTemplate, PromptTemplate
from langchain.agents import AgentExecutor, AgentType, initialize_agent

# agents and tools
from langchain.agents.agent_toolkits.base import BaseToolkit
from langchain.base_language import BaseLanguageModel
from langchain.cache import BaseCache, GPTCache

# callback
from langchain.callbacks.base import BaseCallbackHandler, BaseCallbackManager
from langchain.chains.prompt_selector import ConditionalPromptSelector, is_chat_model
from langchain.chat_models import ChatAnyscale, ChatOpenAI
from langchain.chat_models.base import BaseChatModel
from langchain.docstore.document import Document

# embeddings
from langchain.embeddings.base import Embeddings
from langchain.embeddings import HuggingFaceBgeEmbeddings, HuggingFaceEmbeddings

# LLMs
from langchain.llms import AI21, BaseLLM, Cohere, FakeListLLM, OpenAI
from langchain.memory import ChatMessageHistory, ConversationBufferMemory

# chat and memory
from langchain.memory.chat_memory import BaseChatMemory
from langchain.output_parsers import ResponseSchema
from langchain.prompts.chat import (
    AIMessagePromptTemplate,
    BaseMessagePromptTemplate,
    ChatPromptTemplate,
    HumanMessagePromptTemplate,
    SystemMessagePromptTemplate,
)

# schema
from langchain.schema import (
    AIMessage,
    BaseMemory,
    BaseMessage,
    BaseOutputParser,
    ChatGeneration,
    FunctionMessage,
    HumanMessage,
    LLMResult,
    SystemMessage,
)

# misc
from langchain.sql_database import SQLDatabase
from langchain.input import get_color_mapping, print_text

# input & output
from langchain.text_splitter import RecursiveCharacterTextSplitter, TextSplitter
from langchain.tools import BaseTool, StructuredTool, Tool

__all__ = [
    "langchain",
    "BaseLLM",
    "FakeListLLM",
    "OpenAI",
    "AI21",
    "Cohere",
    "BaseChatModel",
    "ChatAnyscale",
    "ChatOpenAI",
    "BaseLanguageModel",
    "Embeddings",
    "HuggingFaceEmbeddings",
    "HuggingFaceBgeEmbeddings",
    "PromptTemplate",
    "BasePromptTemplate",
    "ConditionalPromptSelector",
    "is_chat_model",
    "AIMessagePromptTemplate",
    "ChatPromptTemplate",
    "HumanMessagePromptTemplate",
    "BaseMessagePromptTemplate",
    "SystemMessagePromptTemplate",
    "BaseChatMemory",
    "ConversationBufferMemory",
    "ChatMessageHistory",
    "BaseToolkit",
    "AgentType",
    "AgentExecutor",
    "initialize_agent",
    "StructuredTool",
    "Tool",
    "BaseTool",
    "ResponseSchema",
    "print_text",
    "get_color_mapping",
    "BaseCallbackHandler",
    "BaseCallbackManager",
    "AIMessage",
    "FunctionMessage",
    "BaseMessage",
    "HumanMessage",
    "SystemMessage",
    "BaseMemory",
    "BaseOutputParser",
    "HumanMessage",
    "BaseMessage",
    "LLMResult",
    "ChatGeneration",
    "SQLDatabase",
    "GPTCache",
    "BaseCache",
    "Document",
    "RecursiveCharacterTextSplitter",
    "TextSplitter",
]



"""Output parsers."""

from llama_index.output_parsers.guardrails import GuardrailsOutputParser
from llama_index.output_parsers.langchain import LangchainOutputParser
from llama_index.output_parsers.pydantic import PydanticOutputParser
from llama_index.output_parsers.selection import SelectionOutputParser

__all__ = [
    "GuardrailsOutputParser",
    "LangchainOutputParser",
    "PydanticOutputParser",
    "SelectionOutputParser",
]



from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, Optional, Type

from llama_index.bridge.pydantic import BaseModel

from llama_index.bridge.langchain import StructuredTool, Tool


class DefaultToolFnSchema(BaseModel):
    """Default tool function Schema."""

    input: str


@dataclass
class ToolMetadata:
    description: str
    name: Optional[str] = None
    fn_schema: Optional[Type[BaseModel]] = DefaultToolFnSchema

    @property
    def fn_schema_str(self) -> str:
        """Get fn schema as string."""
        if self.fn_schema is None:
            raise ValueError("fn_schema is None.")
        return str(self.fn_schema.schema())

    def get_name(self) -> str:
        """Get name."""
        if self.name is None:
            raise ValueError("name is None.")
        return self.name

    def to_openai_function(self) -> Dict[str, Any]:
        """To OpenAI function."""
        if self.fn_schema is None:
            parameters = {
                "properties": {
                    "input": {"title": "input query string", "type": "string"},
                },
                "required": ["input"],
                "type": "object",
            }
        else:
            parameters = self.fn_schema.schema()

        return {
            "name": self.name,
            "description": self.description,
            "parameters": parameters,
        }


class ToolOutput(BaseModel):
    """Tool output."""

    content: str
    tool_name: str
    raw_input: Dict[str, Any]
    raw_output: Any

    def __str__(self) -> str:
        """String."""
        return str(self.content)


class BaseTool:
    @property
    @abstractmethod
    def metadata(self) -> ToolMetadata:
        pass

    @abstractmethod
    def __call__(self, input: Any) -> ToolOutput:
        pass

    def _process_langchain_tool_kwargs(
        self,
        langchain_tool_kwargs: Any,
    ) -> Dict[str, Any]:
        """Process langchain tool kwargs."""
        if "name" not in langchain_tool_kwargs:
            langchain_tool_kwargs["name"] = self.metadata.name or ""
        if "description" not in langchain_tool_kwargs:
            langchain_tool_kwargs["description"] = self.metadata.description
        if "fn_schema" not in langchain_tool_kwargs:
            langchain_tool_kwargs["args_schema"] = self.metadata.fn_schema
        return langchain_tool_kwargs

    def to_langchain_tool(
        self,
        **langchain_tool_kwargs: Any,
    ) -> Tool:
        """To langchain tool."""
        langchain_tool_kwargs = self._process_langchain_tool_kwargs(
            langchain_tool_kwargs
        )
        return Tool.from_function(
            func=self.__call__,
            **langchain_tool_kwargs,
        )

    def to_langchain_structured_tool(
        self,
        **langchain_tool_kwargs: Any,
    ) -> StructuredTool:
        """To langchain structured tool."""
        langchain_tool_kwargs = self._process_langchain_tool_kwargs(
            langchain_tool_kwargs
        )
        return StructuredTool.from_function(
            func=self.__call__,
            **langchain_tool_kwargs,
        )


class AsyncBaseTool(BaseTool):
    """
    Base-level tool class that is backwards compatible with the old tool spec but also
    supports async.
    """

    def __call__(self, *args: Any, **kwargs: Any) -> ToolOutput:
        return self.call(*args, **kwargs)

    @abstractmethod
    def call(self, input: Any) -> ToolOutput:
        """
        This is the method that should be implemented by the tool developer.
        """

    @abstractmethod
    async def acall(self, input: Any) -> ToolOutput:
        """
        This is the async version of the call method.
        Should also be implemented by the tool developer as an
        async-compatible implementation.
        """


class BaseToolAsyncAdapter(AsyncBaseTool):
    """
    Adapter class that allows a synchronous tool to be used as an async tool.
    """

    def __init__(self, tool: BaseTool):
        self.base_tool = tool

    @property
    def metadata(self) -> ToolMetadata:
        return self.base_tool.metadata

    def call(self, input: Any) -> ToolOutput:
        return self.base_tool(input)

    async def acall(self, input: Any) -> ToolOutput:
        return self.call(input)


def adapt_to_async_tool(tool: BaseTool) -> AsyncBaseTool:
    """
    Converts a synchronous tool to an async tool.
    """
    if isinstance(tool, AsyncBaseTool):
        return tool
    else:
        return BaseToolAsyncAdapter(tool)



"""Tools."""

from llama_index.tools.query_engine import QueryEngineTool
from llama_index.tools.retriever_tool import RetrieverTool
from llama_index.tools.types import (
    BaseTool,
    ToolMetadata,
    ToolOutput,
    adapt_to_async_tool,
    AsyncBaseTool,
)
from llama_index.tools.function_tool import FunctionTool
from llama_index.tools.query_plan import QueryPlanTool

__all__ = [
    "BaseTool",
    "adapt_to_async_tool",
    "AsyncBaseTool",
    "QueryEngineTool",
    "RetrieverTool",
    "ToolMetadata",
    "ToolOutput",
    "FunctionTool",
    "QueryPlanTool",
]



import json
from dataclasses import dataclass
from typing import Any, List

from dataclasses_json import DataClassJsonMixin

from llama_index.output_parsers.base import StructuredOutput
from llama_index.output_parsers.utils import _marshal_llm_to_json
from llama_index.types import BaseOutputParser


def _escape_curly_braces(input_string: str) -> str:
    # Replace '{' with '{{' and '}' with '}}' to escape curly braces
    escaped_string = input_string.replace("{", "{{").replace("}", "}}")
    return escaped_string


FORMAT_STR = """The output should be formatted as a JSON instance that conforms to 
the JSON schema below. 

Here is the output schema:
{
  "type": "array",
  "items": {
    "type": "object",
    "properties": {
      "choice": {
        "type": "integer"
      },
      "reason": {
        "type": "string"
      }
    },
    "required": [
      "choice",
      "reason"
    ],
    "additionalProperties": false
  }
}
"""


@dataclass
class Answer(DataClassJsonMixin):
    choice: int
    reason: str


class SelectionOutputParser(BaseOutputParser):
    REQUIRED_KEYS = {"choice", "reason"}

    def _filter_dict(self, json_dict: dict) -> dict:
        output_dict = json_dict
        for key, val in json_dict.items():
            if key in self.REQUIRED_KEYS:
                continue
            elif isinstance(val, dict):
                found = True
                for key in self.REQUIRED_KEYS:
                    if key not in val:
                        found = False
                        break
                if found:
                    output_dict = val
                    break
            elif isinstance(val, list):
                for item in val:
                    if isinstance(item, dict):
                        output_dict = self._filter_dict(item)

        return output_dict

    def _validate_output(self, output: List[dict]) -> List[dict]:
        output_json = []
        for json_dict in output:
            valid = True
            for key in self.REQUIRED_KEYS:
                if key not in json_dict:
                    valid = False
                    break

            if not valid:
                json_dict = self._filter_dict(json_dict)

            output_json.append(json_dict)

        return output_json

    def parse(self, output: str) -> Any:
        output = _marshal_llm_to_json(output)
        json_output = json.loads(output)
        if isinstance(json_output, dict):
            json_output = [json_output]

        json_output = self._validate_output(json_output)
        answers = [Answer.from_dict(json_dict) for json_dict in json_output]
        return StructuredOutput(raw_output=output, parsed_output=answers)

    def format(self, prompt_template: str) -> str:
        fmt = prompt_template + "\n\n" + _escape_curly_braces(FORMAT_STR)
        return fmt



from typing import Dict, Type

from llama_index.vector_stores.chroma import ChromaVectorStore
from llama_index.vector_stores.qdrant import QdrantVectorStore
from llama_index.vector_stores.pinecone import PineconeVectorStore
from llama_index.vector_stores.postgres import PGVectorStore
from llama_index.vector_stores.types import BasePydanticVectorStore
from llama_index.vector_stores.weaviate import WeaviateVectorStore


LOADABLE_VECTOR_STORES: Dict[str, Type[BasePydanticVectorStore]] = {
    ChromaVectorStore.class_name(): ChromaVectorStore,
    QdrantVectorStore.class_name(): QdrantVectorStore,
    PineconeVectorStore.class_name(): PineconeVectorStore,
    PGVectorStore.class_name(): PGVectorStore,
    WeaviateVectorStore.class_name(): WeaviateVectorStore,
}


def load_vector_store(data: dict) -> BasePydanticVectorStore:
    class_name = data.pop("class_name", None)
    if class_name is None:
        raise ValueError("class_name is required to load a vector store")

    if class_name not in LOADABLE_VECTOR_STORES:
        raise ValueError(f"Unable to load vector store of type {class_name}")

    # pop unused keys
    data.pop("flat_metadata", None)
    data.pop("stores_text", None)
    data.pop("is_embedding_query", None)

    if class_name == WeaviateVectorStore.class_name():
        import weaviate

        auth_config_dict = data.pop("auth_config", None)
        if auth_config_dict is not None:
            auth_config = None
            if "api_key" in auth_config_dict:
                auth_config = weaviate.AuthApiKey(**auth_config_dict)
            elif "username" in auth_config_dict:
                auth_config = weaviate.AuthClientPassword(**auth_config_dict)
            else:
                raise ValueError(
                    "Unable to load weaviate auth config, please use an auth "
                    "config with an api_key or username/password."
                )

            data["auth_config"] = auth_config

    return LOADABLE_VECTOR_STORES[class_name].from_params(**data)  # type: ignore



import logging
from typing import Callable, Optional, cast, List

from llama_index.constants import DEFAULT_SIMILARITY_TOP_K
from llama_index.indices.base_retriever import BaseRetriever
from llama_index.indices.query.schema import QueryBundle
from llama_index.indices.vector_store.base import VectorStoreIndex
from llama_index.schema import Document, NodeWithScore
from llama_index.storage.docstore.types import BaseDocumentStore
from llama_index.utils import globals_helper

logger = logging.getLogger(__name__)


class BM25Retriever(BaseRetriever):
    def __init__(
        self,
        docstore: BaseDocumentStore,
        tokenizer: Callable[[str], List[str]],
        similarity_top_k: int = DEFAULT_SIMILARITY_TOP_K,
    ) -> None:
        try:
            from rank_bm25 import BM25Okapi
        except ImportError:
            raise ImportError("Please install rank_bm25: pip install rank-bm25")

        self._docstore = docstore
        self._tokenizer = tokenizer
        self._similarity_top_k = similarity_top_k
        self._documents = cast(
            List[Document], [doc for doc in self._docstore.docs.values()]
        )
        self._corpus = [self._tokenizer(doc.text) for doc in self._documents]

        self.bm25 = BM25Okapi(self._corpus)

    @classmethod
    def from_defaults(
        cls,
        index: VectorStoreIndex,
        tokenizer: Optional[Callable[[str], List[str]]] = None,
        similarity_top_k: int = DEFAULT_SIMILARITY_TOP_K,
    ) -> "BM25Retriever":
        tokenizer = tokenizer or globals_helper.tokenizer
        return cls(
            index.docstore,
            tokenizer,
            similarity_top_k=similarity_top_k,
        )

    def _get_scored_nodes(self, query: str) -> List[NodeWithScore]:
        tokenized_query = self._tokenizer(query)
        doc_scores = self.bm25.get_scores(tokenized_query)

        nodes: List[NodeWithScore] = []
        for i, doc in enumerate(self._documents):
            nodes.append(NodeWithScore(node=doc, score=doc_scores[i]))

        return nodes

    def _retrieve(self, query_bundle: QueryBundle) -> List[NodeWithScore]:
        if query_bundle.custom_embedding_strs or query_bundle.embedding:
            logger.warning("BM25Retriever does not support embeddings, skipping...")

        scored_nodes = self._get_scored_nodes(query_bundle.query_str)

        # Sort and get top_k nodes, score range => 0..1, closer to 1 means more relevant
        nodes = sorted(scored_nodes, key=lambda x: x.score or 0.0, reverse=True)
        top_k_nodes = nodes[: self._similarity_top_k]

        return top_k_nodes



"""Program utils."""

from typing import List, Type

from llama_index.bridge.pydantic import BaseModel, Field, create_model


def create_list_model(base_cls: Type[BaseModel]) -> Type[BaseModel]:
    """Create a list version of an existing Pydantic object."""

    # NOTE: this is directly taken from
    # https://github.com/jxnl/openai_function_call/blob/main/examples/streaming_multitask/streaming_multitask.py
    # all credits go to the openai_function_call repo

    name = f"{base_cls.__name__}List"
    list_items = (
        List[base_cls],  # type: ignore
        Field(
            default_factory=list,
            repr=False,
            description=f"List of {base_cls.__name__} items",
        ),
    )

    new_cls = create_model(name, items=list_items)
    new_cls.__doc__ = f"A list of {base_cls.__name__} objects. "

    return new_cls



from typing import Any, List, Optional, Sequence
from llama_index.indices.base import BaseIndex
from llama_index.indices.composability.graph import ComposableGraph
from llama_index.indices.registry import INDEX_STRUCT_TYPE_TO_INDEX_CLASS
from llama_index.storage.storage_context import StorageContext

import logging

logger = logging.getLogger(__name__)


def load_index_from_storage(
    storage_context: StorageContext,
    index_id: Optional[str] = None,
    **kwargs: Any,
) -> BaseIndex:
    """Load index from storage context.

    Args:
        storage_context (StorageContext): storage context containing
            docstore, index store and vector store.
        index_id (Optional[str]): ID of the index to load.
            Defaults to None, which assumes there's only a single index
            in the index store and load it.
        **kwargs: Additional keyword args to pass to the index constructors.
    """
    index_ids: Optional[Sequence[str]]
    if index_id is None:
        index_ids = None
    else:
        index_ids = [index_id]

    indices = load_indices_from_storage(storage_context, index_ids=index_ids, **kwargs)

    if len(indices) == 0:
        raise ValueError(
            "No index in storage context, check if you specified the right persist_dir."
        )
    elif len(indices) > 1:
        raise ValueError(
            f"Expected to load a single index, but got {len(indices)} instead. "
            "Please specify index_id."
        )

    return indices[0]


def load_indices_from_storage(
    storage_context: StorageContext,
    index_ids: Optional[Sequence[str]] = None,
    **kwargs: Any,
) -> List[BaseIndex]:
    """Load multiple indices from storage context

    Args:
        storage_context (StorageContext): storage context containing
            docstore, index store and vector store.
        index_id (Optional[Sequence[str]]): IDs of the indices to load.
            Defaults to None, which loads all indices in the index store.
        **kwargs: Additional keyword args to pass to the index constructors.
    """
    if index_ids is None:
        logger.info("Loading all indices.")
        index_structs = storage_context.index_store.index_structs()
    else:
        logger.info(f"Loading indices with ids: {index_ids}")
        index_structs = []
        for index_id in index_ids:
            index_struct = storage_context.index_store.get_index_struct(index_id)
            if index_struct is None:
                raise ValueError(f"Failed to load index with ID {index_id}")
            index_structs.append(index_struct)

    indices = []
    for index_struct in index_structs:
        type_ = index_struct.get_type()
        index_cls = INDEX_STRUCT_TYPE_TO_INDEX_CLASS[type_]
        index = index_cls(
            index_struct=index_struct, storage_context=storage_context, **kwargs
        )
        indices.append(index)
    return indices


def load_graph_from_storage(
    storage_context: StorageContext,
    root_id: str,
    **kwargs: Any,
) -> ComposableGraph:
    """Load composable graph from storage context.

    Args:
        storage_context (StorageContext): storage context containing
            docstore, index store and vector store.
        root_id (str): ID of the root index of the graph.
        **kwargs: Additional keyword args to pass to the index constructors.
    """
    indices = load_indices_from_storage(storage_context, index_ids=None, **kwargs)
    all_indices = {index.index_id: index for index in indices}
    graph = ComposableGraph(all_indices=all_indices, root_id=root_id)
    return graph



from typing import Any, Dict, List, Optional, cast

from llama_index.callbacks.base_handler import BaseCallbackHandler
from llama_index.callbacks.schema import CBEventType, EventPayload


class SimpleLLMHandler(BaseCallbackHandler):
    """Callback handler for printing llms inputs/outputs"""

    def __init__(self) -> None:
        super().__init__(event_starts_to_ignore=[], event_ends_to_ignore=[])

    def start_trace(self, trace_id: Optional[str] = None) -> None:
        return

    def end_trace(
        self,
        trace_id: Optional[str] = None,
        trace_map: Optional[Dict[str, List[str]]] = None,
    ) -> None:
        return

    def _print_llm_event(self, payload: dict) -> None:
        from llama_index.llms import ChatMessage

        if EventPayload.PROMPT in payload:
            prompt = str(payload.get(EventPayload.PROMPT))
            completion = str(payload.get(EventPayload.COMPLETION))

            print(f"** Prompt: **\n{prompt}")
            print("*" * 50)
            print(f"** Completion: **\n{completion}")
            print("*" * 50)
            print("\n")
        elif EventPayload.MESSAGES in payload:
            messages = cast(List[ChatMessage], payload.get(EventPayload.MESSAGES, []))
            messages_str = "\n".join([str(x) for x in messages])
            response = str(payload.get(EventPayload.RESPONSE))

            print(f"** Messages: **\n{messages_str}")
            print("*" * 50)
            print(f"** Response: **\n{response}")
            print("*" * 50)
            print("\n")

    def on_event_start(
        self,
        event_type: CBEventType,
        payload: Optional[Dict[str, Any]] = None,
        event_id: str = "",
        **kwargs: Any,
    ) -> str:
        return event_id

    def on_event_end(
        self,
        event_type: CBEventType,
        payload: Optional[Dict[str, Any]] = None,
        event_id: str = "",
        **kwargs: Any,
    ) -> None:
        """Count the LLM or Embedding tokens as needed."""
        if event_type == CBEventType.LLM and payload is not None:
            self._print_llm_event(payload)



from typing import List
import pytest

from llama_index.schema import Document
from llama_index.schema import NodeRelationship, RelatedNodeInfo, TextNode


@pytest.fixture
def documents() -> List[Document]:
    """Get documents."""
    # NOTE: one document for now
    doc_text = (
        "Hello world.\n"
        "This is a test.\n"
        "This is another test.\n"
        "This is a test v2."
    )
    return [Document(text=doc_text)]


@pytest.fixture
def nodes() -> List[TextNode]:
    """Get documents."""
    # NOTE: one document for now
    return [
        TextNode(
            text="Hello world.",
            relationships={
                NodeRelationship.SOURCE: RelatedNodeInfo(node_id="test doc")
            },
        ),
        TextNode(
            text="This is a test.",
            relationships={
                NodeRelationship.SOURCE: RelatedNodeInfo(node_id="test doc")
            },
        ),
        TextNode(
            text="This is another test.",
            relationships={
                NodeRelationship.SOURCE: RelatedNodeInfo(node_id="test doc")
            },
        ),
        TextNode(
            text="This is a test v2.",
            relationships={
                NodeRelationship.SOURCE: RelatedNodeInfo(node_id="test doc")
            },
        ),
    ]



import pytest
from typing import List

from llama_index.bridge.langchain import (
    FakeListLLM,
    SystemMessage,
    HumanMessage,
    AIMessage,
    FunctionMessage,
    BaseMessage,
    Cohere,
    ChatOpenAI,
    OpenAI,
)

from llama_index.llms.base import ChatMessage, MessageRole
from llama_index.llms.langchain import LangChainLLM
from llama_index.llms.langchain_utils import from_lc_messages, to_lc_messages


def test_basic() -> None:
    lc_llm = FakeListLLM(responses=["test response 1", "test response 2"])
    llm = LangChainLLM(llm=lc_llm)

    prompt = "test prompt"
    message = ChatMessage(role="user", content="test message")

    llm.complete(prompt)
    llm.chat([message])


def test_to_lc_messages() -> None:
    lc_messages: List[BaseMessage] = [
        SystemMessage(content="test system message"),
        HumanMessage(content="test human message"),
        AIMessage(content="test ai message"),
        FunctionMessage(content="test function message", name="test function"),
    ]

    messages = from_lc_messages(lc_messages)

    for i in range(len(messages)):
        assert messages[i].content == lc_messages[i].content


def test_from_lc_messages() -> None:
    messages = [
        ChatMessage(content="test system message", role=MessageRole.SYSTEM),
        ChatMessage(content="test human message", role=MessageRole.USER),
        ChatMessage(content="test ai message", role=MessageRole.ASSISTANT),
        ChatMessage(
            content="test function message",
            role=MessageRole.FUNCTION,
            additional_kwargs={"name": "test function"},
        ),
    ]

    lc_messages = to_lc_messages(messages)

    for i in range(len(messages)):
        assert messages[i].content == lc_messages[i].content


try:
    import cohere  # noqa: F401
except ImportError:
    cohere = None  # type: ignore


@pytest.mark.skipif(cohere is None, reason="cohere not installed")
def test_metadata_sets_model_name() -> None:
    chat_gpt = LangChainLLM(
        llm=ChatOpenAI(model="gpt-4-0613", openai_api_key="model-name-tests")
    )
    assert chat_gpt.metadata.model_name == "gpt-4-0613"

    gpt35 = LangChainLLM(
        llm=OpenAI(model="gpt-3.5-turbo-0613", openai_api_key="model-name-tests")
    )
    assert gpt35.metadata.model_name == "gpt-3.5-turbo-0613"

    cohere_llm = LangChainLLM(
        llm=Cohere(model="j2-jumbo-instruct", cohere_api_key="XXXXXXX")
    )
    assert cohere_llm.metadata.model_name == "j2-jumbo-instruct"



from typing import Any, List, Sequence

import pytest
from pytest import MonkeyPatch

from llama_index.agent.openai_agent import OpenAIAgent
from llama_index.chat_engine.types import AgentChatResponse
from llama_index.llms.base import ChatMessage, ChatResponse
from llama_index.llms.mock import MockLLM
from llama_index.llms.openai import OpenAI
from llama_index.tools.function_tool import FunctionTool


def mock_chat_completion(*args: Any, **kwargs: Any) -> dict:
    if "functions" in kwargs:
        if not kwargs["functions"]:
            raise ValueError("functions must not be empty")

    # Example taken from https://platform.openai.com/docs/api-reference/chat/create
    return {
        "id": "chatcmpl-abc123",
        "object": "chat.completion",
        "created": 1677858242,
        "model": "gpt-3.5-turbo-0301",
        "usage": {"prompt_tokens": 13, "completion_tokens": 7, "total_tokens": 20},
        "choices": [
            {
                "message": {"role": "assistant", "content": "\n\nThis is a test!"},
                "finish_reason": "stop",
                "index": 0,
            }
        ],
    }


@pytest.fixture
def add_tool() -> FunctionTool:
    def add(a: int, b: int) -> int:
        """Add two integers and returns the result integer."""
        return a + b

    return FunctionTool.from_defaults(fn=add)


class MockChatLLM(MockLLM):
    def __init__(self, responses: List[ChatMessage]) -> None:
        self._i = 0  # call counter, determines which response to return
        self._responses = responses  # list of responses to return

    def chat(self, messages: Sequence[ChatMessage], **kwargs: Any) -> ChatResponse:
        del messages  # unused
        response = ChatResponse(
            message=self._responses[self._i],
        )
        self._i += 1
        return response


MOCK_ACTION_RESPONSE = """\
Thought: I need to use a tool to help me answer the question.
Action: add
Action Input: {"a": 1, "b": 1}
"""

MOCK_FINAL_RESPONSE = """\
Thought: I have enough information to answer the question without using any more tools.
Answer: 2
"""


def test_chat_basic(
    add_tool: FunctionTool,
    monkeypatch: MonkeyPatch,
) -> None:
    monkeypatch.setattr(
        "llama_index.llms.openai.completion_with_retry", mock_chat_completion
    )

    llm = OpenAI(model="gpt-3.5-turbo")

    agent = OpenAIAgent.from_tools(
        tools=[add_tool],
        llm=llm,
    )
    response = agent.chat("What is 1 + 1?")
    assert isinstance(response, AgentChatResponse)
    assert response.response == "\n\nThis is a test!"


def test_chat_no_functions(
    monkeypatch: MonkeyPatch,
) -> None:
    monkeypatch.setattr(
        "llama_index.llms.openai.completion_with_retry", mock_chat_completion
    )

    llm = OpenAI(model="gpt-3.5-turbo")

    agent = OpenAIAgent.from_tools(
        llm=llm,
    )
    response = agent.chat("What is 1 + 1?")
    assert isinstance(response, AgentChatResponse)
    assert response.response == "\n\nThis is a test!"



from typing import Generator, List

import pytest
from pytest import MonkeyPatch

from llama_index.schema import Document
from llama_index.schema import BaseNode, TextNode
from llama_index.storage.docstore.dynamodb_docstore import DynamoDBDocumentStore
from llama_index.storage.kvstore.dynamodb_kvstore import DynamoDBKVStore

try:
    import boto3
    from moto import mock_dynamodb

    has_boto_libs = True
except ImportError:
    has_boto_libs = False


@pytest.fixture()
def documents() -> List[Document]:
    return [Document(text="doc_1"), Document(text="doc_2")]


@pytest.fixture()
def kvstore_from_mocked_table(
    monkeypatch: MonkeyPatch,
) -> Generator[DynamoDBKVStore, None, None]:
    monkeypatch.setenv("MOTO_ALLOW_NONEXISTENT_REGION", "True")
    monkeypatch.setenv("AWS_DEFAULT_REGION", "Andes")

    table_name = "test_table"
    with mock_dynamodb():
        client = boto3.client("dynamodb")
        client.create_table(
            TableName=table_name,
            AttributeDefinitions=[
                {"AttributeName": "collection", "AttributeType": "S"},
                {"AttributeName": "key", "AttributeType": "S"},
            ],
            KeySchema=[
                {"AttributeName": "collection", "KeyType": "HASH"},
                {"AttributeName": "key", "KeyType": "RANGE"},
            ],
            BillingMode="PAY_PER_REQUEST",
        )
        yield DynamoDBKVStore.from_table_name(table_name)


@pytest.fixture()
def ddb_docstore(kvstore_from_mocked_table: DynamoDBKVStore) -> DynamoDBDocumentStore:
    return DynamoDBDocumentStore(dynamodb_kvstore=kvstore_from_mocked_table)


@pytest.mark.skipif(not has_boto_libs, reason="boto3 and/or moto not installed")
def test_docstore(ddb_docstore: DynamoDBDocumentStore) -> None:
    """Test docstore."""
    doc = Document(text="hello world", id_="d1", metadata={"foo": "bar"})
    node = TextNode(text="my node", id_="d2", metadata={"node": "info"})

    # test get document
    docstore = ddb_docstore
    docstore.add_documents([doc, node])
    gd1 = docstore.get_document("d1")
    assert gd1 == doc
    gd2 = docstore.get_document("d2")
    assert gd2 == node


@pytest.mark.skipif(not has_boto_libs, reason="boto3 and/or moto not installed")
def test_dynamodb_docstore(
    ddb_docstore: DynamoDBDocumentStore, documents: List[Document]
) -> None:
    ds = ddb_docstore
    assert len(ds.docs) == 0

    # test adding documents
    ds.add_documents(documents)
    assert len(ds.docs) == 2
    assert all(isinstance(doc, BaseNode) for doc in ds.docs.values())

    # test updating documents
    ds.add_documents(documents)
    print(ds.docs)
    assert len(ds.docs) == 2

    # test getting documents
    doc0 = ds.get_document(documents[0].get_doc_id())
    assert doc0 is not None
    assert documents[0].get_content() == doc0.get_content()

    # test deleting documents
    ds.delete_document(documents[0].get_doc_id())
    assert len(ds.docs) == 1


@pytest.mark.skipif(not has_boto_libs, reason="boto3 and/or moto not installed")
def test_dynamodb_docstore_hash(
    ddb_docstore: DynamoDBDocumentStore, documents: List[Document]
) -> None:
    ds = ddb_docstore

    # Test setting hash
    ds.set_document_hash("test_doc_id", "test_doc_hash")
    doc_hash = ds.get_document_hash("test_doc_id")
    assert doc_hash == "test_doc_hash"

    # Test updating hash
    ds.set_document_hash("test_doc_id", "test_doc_hash_new")
    doc_hash = ds.get_document_hash("test_doc_id")
    assert doc_hash == "test_doc_hash_new"

    # Test getting non-existent
    doc_hash = ds.get_document_hash("test_not_exist")
    assert doc_hash is None



from abc import ABC, abstractmethod
from typing import List, Optional

from llama_index.indices.query.schema import QueryBundle, QueryType
from llama_index.indices.service_context import ServiceContext
from llama_index.schema import NodeWithScore


class BaseRetriever(ABC):
    """Base retriever."""

    def retrieve(self, str_or_query_bundle: QueryType) -> List[NodeWithScore]:
        """Retrieve nodes given query.

        Args:
            str_or_query_bundle (QueryType): Either a query string or
                a QueryBundle object.

        """
        if isinstance(str_or_query_bundle, str):
            str_or_query_bundle = QueryBundle(str_or_query_bundle)
        return self._retrieve(str_or_query_bundle)

    async def aretrieve(self, str_or_query_bundle: QueryType) -> List[NodeWithScore]:
        if isinstance(str_or_query_bundle, str):
            str_or_query_bundle = QueryBundle(str_or_query_bundle)
        nodes = await self._aretrieve(str_or_query_bundle)
        return nodes

    @abstractmethod
    def _retrieve(self, query_bundle: QueryBundle) -> List[NodeWithScore]:
        """Retrieve nodes given query.

        Implemented by the user.

        """
        pass

    # TODO: make this abstract
    # @abstractmethod
    async def _aretrieve(self, query_bundle: QueryBundle) -> List[NodeWithScore]:
        """Asyncronously retrieve nodes given query.

        Implemented by the user.

        """
        return self._retrieve(query_bundle)

    def get_service_context(self) -> Optional[ServiceContext]:
        """Attempts to resolve a service context.
        Short-circuits at self.service_context, self._service_context,
        or self._index.service_context
        """
        if hasattr(self, "service_context"):
            return self.service_context
        if hasattr(self, "_service_context"):
            return self._service_context
        elif hasattr(self, "_index") and hasattr(self._index, "service_context"):
            return self._index.service_context
        return None



import logging
from threading import Thread
from typing import Any, List, Optional, Union

from llama_index.bridge.pydantic import Field, PrivateAttr

from llama_index.callbacks import CallbackManager
from llama_index.llms.base import (
    CompletionResponse,
    CompletionResponseGen,
    LLMMetadata,
    llm_completion_callback,
)
from llama_index.llms.custom import CustomLLM
from llama_index.prompts.base import PromptTemplate

logger = logging.getLogger(__name__)


class HuggingFaceLLM(CustomLLM):
    """HuggingFace LLM."""

    model_name: str = Field(
        description=(
            "The model name to use from HuggingFace. "
            "Unused if `model` is passed in directly."
        )
    )
    context_window: int = Field(
        description="The maximum number of tokens available for input."
    )
    max_new_tokens: int = Field(description="The maximum number of tokens to generate.")
    system_prompt: str = Field(
        description=(
            "The system prompt, containing any extra instructions or context. "
            "The model card on HuggingFace should specify if this is needed."
        ),
    )
    query_wrapper_prompt: str = Field(
        description=(
            "The query wrapper prompt, containing the query placeholder. "
            "The model card on HuggingFace should specify if this is needed. "
            "Should contain a `{query_str}` placeholder."
        ),
    )
    tokenizer_name: str = Field(
        description=(
            "The name of the tokenizer to use from HuggingFace. "
            "Unused if `tokenizer` is passed in directly."
        )
    )
    device_map: str = Field(description="The device_map to use. Defaults to 'auto'.")
    stopping_ids: List[int] = Field(
        default_factory=list,
        description=(
            "The stopping ids to use. "
            "Generation stops when these token IDs are predicted."
        ),
    )
    tokenizer_outputs_to_remove: list = Field(
        default_factory=list,
        description=(
            "The outputs to remove from the tokenizer. "
            "Sometimes huggingface tokenizers return extra inputs that cause errors."
        ),
    )
    tokenizer_kwargs: dict = Field(
        default_factory=dict, description="The kwargs to pass to the tokenizer."
    )
    model_kwargs: dict = Field(
        default_factory=dict,
        description="The kwargs to pass to the model during initialization.",
    )
    generate_kwargs: dict = Field(
        default_factory=dict,
        description="The kwargs to pass to the model during generation.",
    )

    _model: Any = PrivateAttr()
    _tokenizer: Any = PrivateAttr()
    _stopping_criteria: Any = PrivateAttr()

    def __init__(
        self,
        context_window: int = 4096,
        max_new_tokens: int = 256,
        system_prompt: str = "",
        query_wrapper_prompt: Union[str, PromptTemplate] = "{query_str}",
        tokenizer_name: str = "StabilityAI/stablelm-tuned-alpha-3b",
        model_name: str = "StabilityAI/stablelm-tuned-alpha-3b",
        model: Optional[Any] = None,
        tokenizer: Optional[Any] = None,
        device_map: str = "auto",
        stopping_ids: Optional[List[int]] = None,
        tokenizer_kwargs: Optional[dict] = None,
        tokenizer_outputs_to_remove: Optional[list] = None,
        model_kwargs: Optional[dict] = None,
        generate_kwargs: Optional[dict] = None,
        callback_manager: Optional[CallbackManager] = None,
    ) -> None:
        """Initialize params."""
        import torch
        from transformers import (
            AutoModelForCausalLM,
            AutoTokenizer,
            StoppingCriteria,
            StoppingCriteriaList,
        )

        model_kwargs = model_kwargs or {}
        self._model = model or AutoModelForCausalLM.from_pretrained(
            model_name, device_map=device_map, **model_kwargs
        )

        # check context_window
        config_dict = self._model.config.to_dict()
        model_context_window = int(
            config_dict.get("max_position_embeddings", context_window)
        )
        if model_context_window and model_context_window < context_window:
            logger.warning(
                f"Supplied context_window {context_window} is greater "
                "than the model's max input size {model_context_window}. "
                "Disable this warning by setting a lower context_window."
            )
            context_window = model_context_window

        tokenizer_kwargs = tokenizer_kwargs or {}
        if "max_length" not in tokenizer_kwargs:
            tokenizer_kwargs["max_length"] = context_window

        self._tokenizer = tokenizer or AutoTokenizer.from_pretrained(
            tokenizer_name, **tokenizer_kwargs
        )

        # setup stopping criteria
        stopping_ids_list = stopping_ids or []

        class StopOnTokens(StoppingCriteria):
            def __call__(
                self,
                input_ids: torch.LongTensor,
                scores: torch.FloatTensor,
                **kwargs: Any,
            ) -> bool:
                for stop_id in stopping_ids_list:
                    if input_ids[0][-1] == stop_id:
                        return True
                return False

        self._stopping_criteria = StoppingCriteriaList([StopOnTokens()])

        if isinstance(query_wrapper_prompt, PromptTemplate):
            query_wrapper_prompt = query_wrapper_prompt.template

        super().__init__(
            context_window=context_window,
            max_new_tokens=max_new_tokens,
            system_prompt=system_prompt,
            query_wrapper_prompt=query_wrapper_prompt,
            tokenizer_name=tokenizer_name,
            model_name=model_name,
            device_map=device_map,
            stopping_ids=stopping_ids or [],
            tokenizer_kwargs=tokenizer_kwargs or {},
            tokenizer_outputs_to_remove=tokenizer_outputs_to_remove or [],
            model_kwargs=model_kwargs or {},
            generate_kwargs=generate_kwargs or {},
            callback_manager=callback_manager,
        )

    @classmethod
    def class_name(cls) -> str:
        """Get class name."""
        return "HuggingFace_LLM"

    @property
    def metadata(self) -> LLMMetadata:
        """LLM metadata."""
        return LLMMetadata(
            context_window=self.context_window,
            num_output=self.max_new_tokens,
            model_name=self.model_name,
        )

    @llm_completion_callback()
    def complete(self, prompt: str, **kwargs: Any) -> CompletionResponse:
        """Completion endpoint."""

        full_prompt = prompt
        if self.query_wrapper_prompt:
            full_prompt = self.query_wrapper_prompt.format(query_str=prompt)
        if self.system_prompt:
            full_prompt = f"{self.system_prompt} {full_prompt}"

        inputs = self._tokenizer(full_prompt, return_tensors="pt")
        inputs = inputs.to(self._model.device)

        # remove keys from the tokenizer if needed, to avoid HF errors
        for key in self.tokenizer_outputs_to_remove:
            if key in inputs:
                inputs.pop(key, None)

        tokens = self._model.generate(
            **inputs,
            max_new_tokens=self.max_new_tokens,
            stopping_criteria=self._stopping_criteria,
            **self.generate_kwargs,
        )
        completion_tokens = tokens[0][inputs["input_ids"].size(1) :]
        completion = self._tokenizer.decode(completion_tokens, skip_special_tokens=True)

        return CompletionResponse(text=completion, raw={"model_output": tokens})

    @llm_completion_callback()
    def stream_complete(self, prompt: str, **kwargs: Any) -> CompletionResponseGen:
        """Streaming completion endpoint."""
        from transformers import TextIteratorStreamer

        full_prompt = prompt
        if self.query_wrapper_prompt:
            full_prompt = self.query_wrapper_prompt.format(query_str=prompt)
        if self.system_prompt:
            full_prompt = f"{self.system_prompt} {full_prompt}"

        inputs = self._tokenizer(full_prompt, return_tensors="pt")
        inputs = inputs.to(self._model.device)

        # remove keys from the tokenizer if needed, to avoid HF errors
        for key in self.tokenizer_outputs_to_remove:
            if key in inputs:
                inputs.pop(key, None)

        streamer = TextIteratorStreamer(
            self._tokenizer,
            skip_prompt=True,
            decode_kwargs={"skip_special_tokens": True},
        )
        generation_kwargs = dict(
            inputs,
            streamer=streamer,
            max_new_tokens=self.max_new_tokens,
            stopping_criteria=self._stopping_criteria,
            **self.generate_kwargs,
        )

        # generate in background thread
        # NOTE/TODO: token counting doesn't work with streaming
        thread = Thread(target=self._model.generate, kwargs=generation_kwargs)
        thread.start()

        # create generator based off of streamer
        def gen() -> CompletionResponseGen:
            text = ""
            for x in streamer:
                text += x
                yield CompletionResponse(text=text, delta=x)

        return gen()



"""SQL Structured Store."""
from enum import Enum
from typing import Any, Optional, Sequence, Union

from collections import defaultdict
from llama_index.data_structs.table import SQLStructTable
from llama_index.indices.base_retriever import BaseRetriever
from llama_index.indices.common.struct_store.schema import SQLContextContainer
from llama_index.indices.common.struct_store.sql import SQLStructDatapointExtractor
from llama_index.indices.query.base import BaseQueryEngine
from llama_index.indices.service_context import ServiceContext
from llama_index.indices.struct_store.base import BaseStructStoreIndex
from llama_index.indices.struct_store.container_builder import (
    SQLContextContainerBuilder,
)
from llama_index.langchain_helpers.sql_wrapper import SQLDatabase
from llama_index.schema import BaseNode
from sqlalchemy import Table


class SQLQueryMode(str, Enum):
    SQL = "sql"
    NL = "nl"


class SQLStructStoreIndex(BaseStructStoreIndex[SQLStructTable]):
    """SQL Struct Store Index.

    The SQLStructStoreIndex is an index that uses a SQL database
    under the hood. During index construction, the data can be inferred
    from unstructured documents given a schema extract prompt,
    or it can be pre-loaded in the database.

    During query time, the user can either specify a raw SQL query
    or a natural language query to retrieve their data.

    NOTE: this is deprecated.

    Args:
        documents (Optional[Sequence[DOCUMENTS_INPUT]]): Documents to index.
            NOTE: in the SQL index, this is an optional field.
        sql_database (Optional[SQLDatabase]): SQL database to use,
            including table names to specify.
            See :ref:`Ref-Struct-Store` for more details.
        table_name (Optional[str]): Name of the table to use
            for extracting data.
            Either table_name or table must be specified.
        table (Optional[Table]): SQLAlchemy Table object to use.
            Specifying the Table object explicitly, instead of
            the table name, allows you to pass in a view.
            Either table_name or table must be specified.
        sql_context_container (Optional[SQLContextContainer]): SQL context container.
            an be generated from a SQLContextContainerBuilder.
            See :ref:`Ref-Struct-Store` for more details.

    """

    index_struct_cls = SQLStructTable

    def __init__(
        self,
        nodes: Optional[Sequence[BaseNode]] = None,
        index_struct: Optional[SQLStructTable] = None,
        service_context: Optional[ServiceContext] = None,
        sql_database: Optional[SQLDatabase] = None,
        table_name: Optional[str] = None,
        table: Optional[Table] = None,
        ref_doc_id_column: Optional[str] = None,
        sql_context_container: Optional[SQLContextContainer] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize params."""
        if sql_database is None:
            raise ValueError("sql_database must be specified")
        self.sql_database = sql_database
        # needed here for data extractor
        self._ref_doc_id_column = ref_doc_id_column
        self._table_name = table_name
        self._table = table

        # if documents aren't specified, pass in a blank []
        if index_struct is None:
            nodes = nodes or []

        super().__init__(
            nodes=nodes,
            index_struct=index_struct,
            service_context=service_context,
            **kwargs,
        )

        # TODO: index_struct context_dict is deprecated,
        # we're migrating storage of information to here.
        if sql_context_container is None:
            container_builder = SQLContextContainerBuilder(sql_database)
            sql_context_container = container_builder.build_context_container()
        self.sql_context_container = sql_context_container

    @property
    def ref_doc_id_column(self) -> Optional[str]:
        return self._ref_doc_id_column

    def _build_index_from_nodes(self, nodes: Sequence[BaseNode]) -> SQLStructTable:
        """Build index from nodes."""
        index_struct = self.index_struct_cls()
        if len(nodes) == 0:
            return index_struct
        else:
            data_extractor = SQLStructDatapointExtractor(
                self._service_context.llm_predictor,
                self.schema_extract_prompt,
                self.output_parser,
                self.sql_database,
                table_name=self._table_name,
                table=self._table,
                ref_doc_id_column=self._ref_doc_id_column,
            )
            # group nodes by ids
            source_to_node = defaultdict(list)
            for node in nodes:
                source_to_node[node.ref_doc_id].append(node)

            for _, node_set in source_to_node.items():
                data_extractor.insert_datapoint_from_nodes(node_set)
        return index_struct

    def _insert(self, nodes: Sequence[BaseNode], **insert_kwargs: Any) -> None:
        """Insert a document."""
        data_extractor = SQLStructDatapointExtractor(
            self._service_context.llm_predictor,
            self.schema_extract_prompt,
            self.output_parser,
            self.sql_database,
            table_name=self._table_name,
            table=self._table,
            ref_doc_id_column=self._ref_doc_id_column,
        )
        data_extractor.insert_datapoint_from_nodes(nodes)

    def as_retriever(self, **kwargs: Any) -> BaseRetriever:
        raise NotImplementedError("Not supported")

    def as_query_engine(
        self, query_mode: Union[str, SQLQueryMode] = SQLQueryMode.NL, **kwargs: Any
    ) -> BaseQueryEngine:
        # NOTE: lazy import
        from llama_index.indices.struct_store.sql_query import (
            NLStructStoreQueryEngine,
            SQLStructStoreQueryEngine,
        )

        if query_mode == SQLQueryMode.NL:
            return NLStructStoreQueryEngine(self, **kwargs)
        elif query_mode == SQLQueryMode.SQL:
            return SQLStructStoreQueryEngine(self, **kwargs)
        else:
            raise ValueError(f"Unknown query mode: {query_mode}")


GPTSQLStructStoreIndex = SQLStructStoreIndex



from __future__ import annotations

from typing import Any, Optional, Tuple, cast

from llama_index.bridge.pydantic import PrivateAttr

from llama_index.prompts import BasePromptTemplate
from llama_index.callbacks import CallbackManager
from llama_index.callbacks.schema import CBEventType, EventPayload
from llama_index.llm_predictor.base import LLM, BaseLLMPredictor, LLMMetadata
from llama_index.llm_predictor.vellum.exceptions import VellumGenerateException
from llama_index.llm_predictor.vellum.prompt_registry import VellumPromptRegistry
from llama_index.llm_predictor.vellum.types import (
    VellumCompiledPrompt,
    VellumRegisteredPrompt,
)
from llama_index.types import TokenAsyncGen, TokenGen


class VellumPredictor(BaseLLMPredictor):
    _callback_manager: CallbackManager = PrivateAttr(default_factory=CallbackManager)

    _vellum_client: Any = PrivateAttr()
    _async_vellum_client = PrivateAttr()
    _prompt_registry: Any = PrivateAttr()

    class Config:
        arbitrary_types_allowed = True

    def __init__(
        self,
        vellum_api_key: str,
        callback_manager: Optional[CallbackManager] = None,
    ) -> None:
        import_err_msg = (
            "`vellum` package not found, please run `pip install vellum-ai`"
        )
        try:
            from vellum.client import AsyncVellum, Vellum  # noqa: F401
        except ImportError:
            raise ImportError(import_err_msg)

        self._callback_manager = callback_manager or CallbackManager([])

        # Vellum-specific
        self._vellum_client = Vellum(api_key=vellum_api_key)
        self._async_vellum_client = AsyncVellum(api_key=vellum_api_key)
        self._prompt_registry = VellumPromptRegistry(vellum_api_key=vellum_api_key)

        super().__init__()

    @classmethod
    def class_name(cls) -> str:
        """Get class name."""
        return "VellumPredictor"

    @property
    def metadata(self) -> LLMMetadata:
        """Get LLM metadata."""

        # Note: We use default values here, but ideally we would retrieve this metadata
        # via Vellum's API based on the LLM that backs the registered prompt's
        # deployment. This is not currently possible, so we use default values.
        return LLMMetadata()

    @property
    def callback_manager(self) -> CallbackManager:
        """Get callback manager."""
        return self._callback_manager

    @property
    def llm(self) -> LLM:
        """Get the LLM."""
        raise NotImplementedError("Vellum does not expose the LLM.")

    def predict(self, prompt: BasePromptTemplate, **prompt_args: Any) -> str:
        """Predict the answer to a query."""

        from vellum import GenerateRequest

        registered_prompt, compiled_prompt, event_id = self._prepare_generate_call(
            prompt, **prompt_args
        )

        input_values = {
            **prompt.kwargs,
            **prompt_args,
        }
        result = self._vellum_client.generate(
            deployment_id=registered_prompt.deployment_id,
            requests=[GenerateRequest(input_values=input_values)],
        )

        completion_text = self._process_generate_response(
            result, compiled_prompt, event_id
        )

        return completion_text

    def stream(self, prompt: BasePromptTemplate, **prompt_args: Any) -> TokenGen:
        """Stream the answer to a query."""

        from vellum import GenerateRequest, GenerateStreamResult

        registered_prompt, compiled_prompt, event_id = self._prepare_generate_call(
            prompt, **prompt_args
        )

        input_values = {
            **prompt.kwargs,
            **prompt_args,
        }
        responses = self._vellum_client.generate_stream(
            deployment_id=registered_prompt.deployment_id,
            requests=[GenerateRequest(input_values=input_values)],
        )

        def text_generator() -> TokenGen:
            complete_text = ""

            while True:
                try:
                    stream_response = next(responses)
                except StopIteration:
                    self.callback_manager.on_event_end(
                        CBEventType.LLM,
                        payload={
                            EventPayload.RESPONSE: complete_text,
                            EventPayload.PROMPT: compiled_prompt.text,
                        },
                        event_id=event_id,
                    )
                    break

                result: GenerateStreamResult = stream_response.delta

                if result.error:
                    raise VellumGenerateException(result.error.message)
                elif not result.data:
                    raise VellumGenerateException(
                        "Unknown error occurred while generating"
                    )

                completion_text_delta = result.data.completion.text
                complete_text += completion_text_delta

                yield completion_text_delta

        return text_generator()

    async def apredict(self, prompt: BasePromptTemplate, **prompt_args: Any) -> str:
        """Asynchronously predict the answer to a query."""

        from vellum import GenerateRequest

        registered_prompt, compiled_prompt, event_id = self._prepare_generate_call(
            prompt, **prompt_args
        )

        input_values = {
            **prompt.kwargs,
            **prompt_args,
        }
        result = await self._async_vellum_client.generate(
            deployment_id=registered_prompt.deployment_id,
            requests=[GenerateRequest(input_values=input_values)],
        )

        completion_text = self._process_generate_response(
            result, compiled_prompt, event_id
        )

        return completion_text

    async def astream(
        self, prompt: BasePromptTemplate, **prompt_args: Any
    ) -> TokenAsyncGen:
        async def gen() -> TokenAsyncGen:
            for token in self.stream(prompt, **prompt_args):
                yield token

        # NOTE: convert generator to async generator
        return gen()

    def _prepare_generate_call(
        self, prompt: BasePromptTemplate, **prompt_args: Any
    ) -> Tuple[VellumRegisteredPrompt, VellumCompiledPrompt, str]:
        """Prepare a generate call."""

        registered_prompt = self._prompt_registry.from_prompt(prompt)
        compiled_prompt = self._prompt_registry.get_compiled_prompt(
            registered_prompt, prompt_args
        )

        cb_payload = {
            **prompt_args,
            "deployment_id": registered_prompt.deployment_id,
            "model_version_id": registered_prompt.model_version_id,
        }
        event_id = self.callback_manager.on_event_start(
            CBEventType.LLM,
            payload=cb_payload,
        )
        return registered_prompt, compiled_prompt, event_id

    def _process_generate_response(
        self,
        result: Any,
        compiled_prompt: VellumCompiledPrompt,
        event_id: str,
    ) -> str:
        """Process the response from a generate call."""

        from vellum import GenerateResponse

        result = cast(GenerateResponse, result)

        completion_text = result.text

        self.callback_manager.on_event_end(
            CBEventType.LLM,
            payload={
                EventPayload.RESPONSE: completion_text,
                EventPayload.PROMPT: compiled_prompt.text,
            },
            event_id=event_id,
        )

        return completion_text



"""Sentence Transformer Finetuning Engine."""

from llama_index.embeddings.base import BaseEmbedding

from typing import Any, Optional

from llama_index.embeddings.utils import resolve_embed_model
from llama_index.finetuning.types import BaseEmbeddingFinetuneEngine
from llama_index.finetuning.embeddings.common import (
    EmbeddingQAFinetuneDataset,
)


class SentenceTransformersFinetuneEngine(BaseEmbeddingFinetuneEngine):
    """Sentence Transformers Finetune Engine."""

    def __init__(
        self,
        dataset: EmbeddingQAFinetuneDataset,
        model_id: str = "BAAI/bge-small-en",
        model_output_path: str = "exp_finetune",
        batch_size: int = 10,
        val_dataset: Optional[EmbeddingQAFinetuneDataset] = None,
        loss: Optional[Any] = None,
        epochs: int = 2,
        show_progress_bar: bool = True,
        evaluation_steps: int = 50,
    ) -> None:
        """Init params."""
        from sentence_transformers import InputExample, SentenceTransformer, losses
        from torch.utils.data import DataLoader

        self.dataset = dataset

        self.model_id = model_id
        self.model_output_path = model_output_path
        self.model = SentenceTransformer(model_id)

        # TODO: support more than 1 doc per query
        examples: Any = []
        for query_id, query in dataset.queries.items():
            node_id = dataset.relevant_docs[query_id][0]
            text = dataset.corpus[node_id]
            example = InputExample(texts=[query, text])
            examples.append(example)
        self.examples = examples

        self.loader: DataLoader = DataLoader(examples, batch_size=batch_size)

        # define evaluator
        from sentence_transformers.evaluation import InformationRetrievalEvaluator

        evaluator: Optional[InformationRetrievalEvaluator] = None
        if val_dataset is not None:
            evaluator = InformationRetrievalEvaluator(
                val_dataset.queries, val_dataset.corpus, val_dataset.relevant_docs
            )
        self.evaluator = evaluator

        # define loss
        self.loss = loss or losses.MultipleNegativesRankingLoss(self.model)

        self.epochs = epochs
        self.show_progress_bar = show_progress_bar
        self.evaluation_steps = evaluation_steps
        self.warmup_steps = int(len(self.loader) * epochs * 0.1)

    def finetune(self, **train_kwargs: Any) -> None:
        """Finetune model."""
        self.model.fit(
            train_objectives=[(self.loader, self.loss)],
            epochs=self.epochs,
            warmup_steps=self.warmup_steps,
            output_path=self.model_output_path,
            show_progress_bar=self.show_progress_bar,
            evaluator=self.evaluator,
            evaluation_steps=self.evaluation_steps,
        )

    def get_finetuned_model(self, **model_kwargs: Any) -> BaseEmbedding:
        """Gets finetuned model."""

        embed_model_str = "local:" + self.model_output_path
        embed_model = resolve_embed_model(embed_model_str)

        return embed_model



from typing import Any, Dict, Optional
from llama_index.storage.kvstore.types import DEFAULT_COLLECTION, BaseKVStore

# keyword "_" is reserved in Firestore but refered in llama_index/constants.py.
FIELD_NAME_REPLACE_SET = {"__data__": "data", "__type__": "type"}
FIELD_NAME_REPLACE_GET = {"data": "__data__", "type": "__type__"}

# "/" is not supported in Firestore Collection ID.
SLASH_REPLACEMENT = "_"
IMPORT_ERROR_MSG = (
    "`firestore` package not found, please run `pip3 install google-cloud-firestore`"
)
CLIENT_INFO = "LlamaIndex"
DEFAULT_FIRESTORE_DATABASE = "(default)"


class FirestoreKVStore(BaseKVStore):
    """Firestore Key-Value store.

    Args:
        project (str): The project which the client acts on behalf of.
        database (str): The database name that the client targets.
    """

    def __init__(
        self, project: Optional[str] = None, database: str = DEFAULT_FIRESTORE_DATABASE
    ) -> None:
        try:
            from google.cloud import firestore_v1 as firestore
        except ImportError:
            raise ImportError(IMPORT_ERROR_MSG)

        self._db = firestore.client.Client(
            project=project, database=database, client_info=CLIENT_INFO
        )

    def firestore_collection(self, collection: str) -> str:
        return collection.replace("/", SLASH_REPLACEMENT)

    def replace_field_name_set(self, val: Dict[str, Any]) -> Dict[str, Any]:
        val = val.copy()
        for k, v in FIELD_NAME_REPLACE_SET.items():
            if k in val:
                val[v] = val[k]
                val.pop(k)
        return val

    def replace_field_name_get(self, val: Dict[str, Any]) -> Dict[str, Any]:
        val = val.copy()
        for k, v in FIELD_NAME_REPLACE_GET.items():
            if k in val:
                val[v] = val[k]
                val.pop(k)
        return val

    def put(
        self,
        key: str,
        val: dict,
        collection: str = DEFAULT_COLLECTION,
    ) -> None:
        """Put a key-value pair into the Firestore collection.
        Args:
            key (str): key
            val (dict): value
            collection (str): collection name
        """
        collection_id = self.firestore_collection(collection)
        val = self.replace_field_name_set(val)
        doc = self._db.collection(collection_id).document(key)
        doc.set(val, merge=True)

    def get(self, key: str, collection: str = DEFAULT_COLLECTION) -> Optional[dict]:
        """Get a key-value pair from the Firestore.
        Args:
            key (str): key
            collection (str): collection name
        """
        collection_id = self.firestore_collection(collection)
        result = self._db.collection(collection_id).document(key).get().to_dict()
        if not result:
            return None

        return self.replace_field_name_get(result)

    def get_all(self, collection: str = DEFAULT_COLLECTION) -> Dict[str, dict]:
        """Get all values from the Firestore collection.
        Args:
            collection (str): collection name
        """
        collection_id = self.firestore_collection(collection)
        docs = self._db.collection(collection_id).list_documents()
        output = {}
        for doc in docs:
            key = doc.id
            val = self.replace_field_name_get(doc.get().to_dict())
            output[key] = val
        return output

    def delete(self, key: str, collection: str = DEFAULT_COLLECTION) -> bool:
        """Delete a value from the Firestore.
        Args:
            key (str): key
            collection (str): collection name
        """
        collection_id = self.firestore_collection(collection)
        doc = self._db.collection(collection_id).document(key)
        doc.delete()
        return True



from typing import Any, Dict, List, Optional

from llama_index.schema import BaseNode, MetadataMode
from llama_index.vector_stores.types import (
    VectorStore,
    VectorStoreQuery,
    VectorStoreQueryResult,
)
from llama_index.vector_stores.utils import metadata_dict_to_node, node_to_metadata_dict


def check_if_not_null(props: List[str], values: List[Any]) -> None:
    """Check if variable is not null and raise error accordingly"""
    for prop, value in zip(props, values):
        if not value:
            raise ValueError(f"Parameter `{prop}` must not be None or empty string")


def sort_by_index_name(
    lst: List[Dict[str, Any]], index_name: str
) -> List[Dict[str, Any]]:
    """Sort first element to match the index_name if exists"""
    return sorted(lst, key=lambda x: x.get("index_name") != index_name)


def clean_params(params: List[BaseNode]) -> List[Dict[str, Any]]:
    """Convert BaseNode object to a dictionary to be imported into Neo4j"""
    clean_params = []
    for record in params:
        text = record.get_content(metadata_mode=MetadataMode.NONE)
        embedding = record.get_embedding()
        id = record.node_id
        metadata = node_to_metadata_dict(record, remove_text=True, flat_metadata=False)
        # Remove redundant metadata information
        for k in ["document_id", "doc_id"]:
            del metadata[k]
        clean_params.append(
            {"text": text, "embedding": embedding, "id": id, "metadata": metadata}
        )
    return clean_params


class Neo4jVectorStore(VectorStore):
    stores_text: bool = True
    flat_metadata = True

    def __init__(
        self,
        username: str,
        password: str,
        url: str,
        embedding_dimension: int,
        database: str = "neo4j",
        index_name: str = "vector",
        node_label: str = "Chunk",
        embedding_node_property: str = "embedding",
        text_node_property: str = "text",
        distance_strategy: str = "cosine",
        retrieval_query: str = "",
        **kwargs: Any,
    ) -> None:
        try:
            import neo4j
        except ImportError:
            raise ImportError(
                "Could not import neo4j python package. "
                "Please install it with `pip install neo4j`."
            )
        if distance_strategy not in ["cosine", "euclidean"]:
            raise ValueError("distance_strategy must be either 'euclidean' or 'cosine'")

        self._driver = neo4j.GraphDatabase.driver(url, auth=(username, password))
        self._database = database

        # Verify connection
        try:
            self._driver.verify_connectivity()
        except neo4j.exceptions.ServiceUnavailable:
            raise ValueError(
                "Could not connect to Neo4j database. "
                "Please ensure that the url is correct"
            )
        except neo4j.exceptions.AuthError:
            raise ValueError(
                "Could not connect to Neo4j database. "
                "Please ensure that the username and password are correct"
            )

        # Verify if the version support vector index
        self._verify_version()

        # Verify that required values are not null
        check_if_not_null(
            [
                "index_name",
                "node_label",
                "embedding_node_property",
                "text_node_property",
            ],
            [index_name, node_label, embedding_node_property, text_node_property],
        )

        self.distance_strategy = distance_strategy
        self.index_name = index_name
        self.node_label = node_label
        self.embedding_node_property = embedding_node_property
        self.text_node_property = text_node_property
        self.retrieval_query = retrieval_query
        self.embedding_dimension = embedding_dimension

        index_already_exists = self.retrieve_existing_index()
        if not index_already_exists:
            self.create_new_index()

    def _verify_version(self) -> None:
        """
        Check if the connected Neo4j database version supports vector indexing.

        Queries the Neo4j database to retrieve its version and compares it
        against a target version (5.11.0) that is known to support vector
        indexing. Raises a ValueError if the connected Neo4j version is
        not supported.
        """
        version = self.database_query("CALL dbms.components()")[0]["versions"][0]
        if "aura" in version:
            version_tuple = tuple(map(int, version.split("-")[0].split("."))) + (0,)
        else:
            version_tuple = tuple(map(int, version.split(".")))

        target_version = (5, 11, 0)

        if version_tuple < target_version:
            raise ValueError(
                "Version index is only supported in Neo4j version 5.11 or greater"
            )

    def create_new_index(self) -> None:
        """
        This method constructs a Cypher query and executes it
        to create a new vector index in Neo4j.
        """
        index_query = (
            "CALL db.index.vector.createNodeIndex("
            "$index_name,"
            "$node_label,"
            "$embedding_node_property,"
            "toInteger($embedding_dimension),"
            "$similarity_metric )"
        )

        parameters = {
            "index_name": self.index_name,
            "node_label": self.node_label,
            "embedding_node_property": self.embedding_node_property,
            "embedding_dimension": self.embedding_dimension,
            "similarity_metric": self.distance_strategy,
        }
        self.database_query(index_query, params=parameters)

    def retrieve_existing_index(self) -> bool:
        """
        Check if the vector index exists in the Neo4j database
        and returns its embedding dimension.

        This method queries the Neo4j database for existing indexes
        and attempts to retrieve the dimension of the vector index
        with the specified name. If the index exists, its dimension is returned.
        If the index doesn't exist, `None` is returned.

        Returns:
            int or None: The embedding dimension of the existing index if found.
        """

        index_information = self.database_query(
            "SHOW INDEXES YIELD name, type, labelsOrTypes, properties, options "
            "WHERE type = 'VECTOR' AND (name = $index_name "
            "OR (labelsOrTypes[0] = $node_label AND "
            "properties[0] = $embedding_node_property)) "
            "RETURN name, labelsOrTypes, properties, options ",
            params={
                "index_name": self.index_name,
                "node_label": self.node_label,
                "embedding_node_property": self.embedding_node_property,
            },
        )
        # sort by index_name
        index_information = sort_by_index_name(index_information, self.index_name)
        try:
            self.index_name = index_information[0]["name"]
            self.node_label = index_information[0]["labelsOrTypes"][0]
            self.embedding_node_property = index_information[0]["properties"][0]
            self.embedding_dimension = index_information[0]["options"]["indexConfig"][
                "vector.dimensions"
            ]

            return True
        except IndexError:
            return False

    def database_query(
        self, query: str, params: Optional[dict] = None
    ) -> List[Dict[str, Any]]:
        """
        This method sends a Cypher query to the connected Neo4j database
        and returns the results as a list of dictionaries.

        Args:
            query (str): The Cypher query to execute.
            params (dict, optional): Dictionary of query parameters. Defaults to {}.

        Returns:
            List[Dict[str, Any]]: List of dictionaries containing the query results.
        """
        from neo4j.exceptions import CypherSyntaxError

        params = params or {}
        with self._driver.session(database=self._database) as session:
            try:
                data = session.run(query, params)
                return [r.data() for r in data]
            except CypherSyntaxError as e:
                raise ValueError(f"Cypher Statement is not valid\n{e}")

    def add(self, nodes: List[BaseNode]) -> List[str]:
        ids = [r.node_id for r in nodes]
        import_query = (
            "UNWIND $data AS row "
            "CALL { WITH row "
            f"MERGE (c:`{self.node_label}` {{id: row.id}}) "
            "WITH c, row "
            f"CALL db.create.setVectorProperty(c, "
            f"'{self.embedding_node_property}', row.embedding) "
            "YIELD node "
            f"SET c.`{self.text_node_property}` = row.text "
            "SET c += row.metadata } IN TRANSACTIONS OF 1000 ROWS"
        )

        self.database_query(
            import_query,
            params={"data": clean_params(nodes)},
        )

        return ids

    def query(self, query: VectorStoreQuery, **kwargs: Any) -> VectorStoreQueryResult:
        default_retrieval = (
            f"RETURN node.`{self.text_node_property}` AS text, score, "
            "node.id AS id, "
            f"node {{.*, `{self.text_node_property}`: Null, "
            f"`{self.embedding_node_property}`: Null, id: Null }} AS metadata"
        )

        retrieval_query = self.retrieval_query or default_retrieval

        read_query = (
            "CALL db.index.vector.queryNodes($index, $k, $embedding) "
            "YIELD node, score "
        ) + retrieval_query

        parameters = {
            "index": self.index_name,
            "k": query.similarity_top_k,
            "embedding": query.query_embedding,
        }

        results = self.database_query(read_query, params=parameters)

        nodes = []
        similarities = []
        ids = []
        for record in results:
            node = metadata_dict_to_node(record["metadata"])
            node.set_content(str(record["text"]))
            nodes.append(node)
            similarities.append(record["score"])
            ids.append(record["id"])

        return VectorStoreQueryResult(nodes=nodes, similarities=similarities, ids=ids)

    def delete(self, ref_doc_id: str, **delete_kwargs: Any) -> None:
        self.database_query(
            f"MATCH (n:`{self.node_label}`) WHERE n.ref_doc_id = $id DETACH DELETE n",
            params={"id": ref_doc_id},
        )



from typing import List, Any, Generator, Dict
import pytest
from unittest.mock import patch, MagicMock
from llama_index.llms.base import (
    ChatMessage,
    MessageRole,
)
from llama_index.llms.rungpt import RunGptLLM

try:
    import sseclient
except ImportError:
    sseclient = None


def mock_completion(*args: Any, **kwargs: Any) -> Dict[str, Any]:
    # Example taken from rungpt example inferece code on github repo.
    return {
        "id": None,
        "object": "text_completion",
        "created": 1692891018,
        "choices": [
            {"text": "This is an indeed test.", "finish_reason": "length", "index": 0.0}
        ],
        "prompt": "Once upon a time,",
        "usage": {"completion_tokens": 21, "total_tokens": 27, "prompt_tokens": 6},
    }


def mock_chat_completion(*args: Any, **kwargs: Any) -> Dict[str, Any]:
    # Example taken from rungpt example inferece code on github repo.
    return {
        "id": None,
        "object": "chat.completion",
        "created": 1692892252,
        "choices": [
            {
                "finish_reason": "length",
                "index": 0.0,
                "message": {"content": "This is an indeed test.", "role": "assistant"},
            }
        ],
        "prompt": "Test prompt",
        "usage": {"completion_tokens": 59, "total_tokens": 103, "prompt_tokens": 44},
    }


def mock_completion_stream(*args: Any, **kwargs: Any) -> Generator[str, None, None]:
    # Example taken from rungpt example inferece code on github repo.
    events = [
        str(
            {
                "id": None,
                "object": "text_completion",
                "created": 1692891964,
                "choices": [{"text": "This", "finish_reason": None, "index": 0.0}],
                "prompt": "This",
                "usage": {
                    "completion_tokens": 1,
                    "total_tokens": 7,
                    "prompt_tokens": 6,
                },
            }
        ),
        str(
            {
                "id": None,
                "object": "text_completion",
                "created": 1692891964,
                "choices": [{"text": " is", "finish_reason": None, "index": 0.0}],
                "prompt": " is",
                "usage": {
                    "completion_tokens": 2,
                    "total_tokens": 9,
                    "prompt_tokens": 7,
                },
            }
        ),
        str(
            {
                "id": None,
                "object": "text_completion",
                "created": 1692891964,
                "choices": [{"text": " test.", "finish_reason": None, "index": 0.0}],
                "prompt": " test.",
                "usage": {
                    "completion_tokens": 3,
                    "total_tokens": 11,
                    "prompt_tokens": 8,
                },
            }
        ),
    ]
    for event in events:
        yield event


def mock_chat_completion_stream(
    *args: Any, **kwargs: Any
) -> Generator[str, None, None]:
    # Example taken from rungpt example inferece code on github repo.
    events = [
        str(
            {
                "id": None,
                "object": "chat.completion",
                "created": 1692892378,
                "choices": [
                    {
                        "finish_reason": None,
                        "index": 0.0,
                        "message": {"content": "This", "role": "assistant"},
                    }
                ],
                "prompt": "Mock prompt",
                "usage": {
                    "completion_tokens": 1,
                    "total_tokens": 45,
                    "prompt_tokens": 44,
                },
            }
        ),
        str(
            {
                "id": None,
                "object": "chat.completion",
                "created": 1692892378,
                "choices": [
                    {
                        "finish_reason": None,
                        "index": 0.0,
                        "message": {"content": " is", "role": "assistant"},
                    }
                ],
                "prompt": None,
                "usage": {
                    "completion_tokens": 2,
                    "total_tokens": 47,
                    "prompt_tokens": 45,
                },
            }
        ),
        str(
            {
                "id": None,
                "object": "chat.completion",
                "created": 1692892379,
                "choices": [
                    {
                        "finish_reason": None,
                        "index": 0.0,
                        "message": {"content": " test.", "role": "assistant"},
                    }
                ],
                "prompt": None,
                "usage": {
                    "completion_tokens": 3,
                    "total_tokens": 49,
                    "prompt_tokens": 46,
                },
            }
        ),
    ]
    for event in events:
        yield event


def mock_chat_history(*args: Any, **kwargs: Any) -> List[ChatMessage]:
    return [
        ChatMessage(
            role=MessageRole.USER,
            message="Hello, my name is zihao, major in artificial intelligence.",
        ),
        ChatMessage(
            role=MessageRole.ASSISTANT,
            message="Hello, what can I do for you?",
        ),
        ChatMessage(
            role=MessageRole.USER,
            message="Could you tell me what is my name and major?",
        ),
    ]


def test_init() -> None:
    dummy = RunGptLLM(model="mock model", endpoint="0.0.0.0:51002")
    assert dummy.model == "mock model"
    assert dummy.endpoint == "0.0.0.0:51002"
    assert isinstance(dummy, RunGptLLM)


def test_complete() -> None:
    dummy = RunGptLLM()
    with patch("requests.post") as mock_post:
        mock_post.return_value.json.return_value = mock_completion()
        response = dummy.complete("mock prompt")
        assert response.text == "This is an indeed test."


@pytest.mark.parametrize(
    "chat_history", [mock_chat_history(), tuple(mock_chat_history())]
)
def test_chat(chat_history: List[ChatMessage]) -> None:
    with patch("requests.post") as mock_post:
        mock_post.return_value.json.return_value = mock_chat_completion()
        dummy = RunGptLLM()
        response = dummy.chat(chat_history)
        assert response.message.content == "This is an indeed test."
        assert response.message.role == "assistant"


@pytest.mark.skipif(sseclient is None, reason="sseclient not installed")
@pytest.mark.parametrize(
    "chat_history", [mock_chat_history(), tuple(mock_chat_history())]
)
def test_stream_chat(chat_history: List[ChatMessage]) -> None:
    mock_events = [
        MagicMock(data=event_data) for event_data in mock_chat_completion_stream()
    ]
    mock_event_iterator = iter(mock_events)

    with patch("requests.post"), patch("sseclient.SSEClient") as mock_sseclient:
        mock_response = MagicMock()
        mock_response.json.return_value = {}
        type(mock_response).status_code = 200
        mock_sseclient.return_value.events.return_value = mock_event_iterator

        dummy = RunGptLLM()
        response_gen = dummy.stream_chat(chat_history)
        responses = list(response_gen)
        assert responses[-1].message.content == " This is test."
        assert responses[-1].message.role == "assistant"


@pytest.mark.skipif(sseclient is None, reason="sseclient not installed")
def test_stream_complete() -> None:
    mock_events = [
        MagicMock(data=event_data) for event_data in mock_completion_stream()
    ]
    mock_event_iterator = iter(mock_events)
    mock_prompt = "A mock prompt"

    with patch("requests.post"), patch("sseclient.SSEClient") as mock_sseclient:
        mock_response = MagicMock()
        mock_response.json.return_value = {}
        type(mock_response).status_code = 200
        mock_sseclient.return_value.events.return_value = mock_event_iterator

        dummy = RunGptLLM()
        response_gen = dummy.stream_complete(mock_prompt)
        responses = list(response_gen)
        assert responses[-1].text == " This is test."
        assert responses[-1].delta == " test."



from typing import Any, Sequence
from llama_index.response_synthesizers.base import BaseSynthesizer
from llama_index.types import RESPONSE_TEXT_TYPE


class NoText(BaseSynthesizer):
    def get_response(
        self,
        query_str: str,
        text_chunks: Sequence[str],
        **response_kwargs: Any,
    ) -> RESPONSE_TEXT_TYPE:
        return ""

    async def aget_response(
        self,
        query_str: str,
        text_chunks: Sequence[str],
        **response_kwargs: Any,
    ) -> RESPONSE_TEXT_TYPE:
        return ""



"""Test finetuning engine."""
import pkgutil
import pytest


def test_torch_imports() -> None:
    """Test that torch is an optional dependency."""
    # importing fine-tuning modules should be ok
    from llama_index.finetuning import OpenAIFinetuneEngine  # noqa: F401
    from llama_index.finetuning import EmbeddingAdapterFinetuneEngine  # noqa: F401
    from llama_index.finetuning import SentenceTransformersFinetuneEngine  # noqa: F401

    # if torch isn't installed, then these should fail
    if pkgutil.find_loader("torch") is None:
        with pytest.raises(ModuleNotFoundError):
            from llama_index.embeddings.adapter_utils import LinearLayer  # noqa: F401
            from llama_index.finetuning.embeddings.adapter_utils import (
                train_model,
            )  # noqa: F401
    else:
        # else, importing these should be ok
        from llama_index.embeddings.adapter_utils import LinearLayer  # noqa: F401
        from llama_index.finetuning.embeddings.adapter_utils import (
            train_model,  # noqa: F401
        )



from __future__ import annotations

from decimal import Decimal
from typing import Any, Dict, List, Optional, Set, Tuple

from llama_index.storage.kvstore.types import DEFAULT_COLLECTION, BaseKVStore
import os

IMPORT_ERROR_MSG = "`boto3` package not found, please run `pip install boto3`"


def parse_schema(table: Any) -> Tuple[str, str]:
    key_hash: Optional[str] = None
    key_range: Optional[str] = None

    for key in table.key_schema:
        if key["KeyType"] == "HASH":
            key_hash = key["AttributeName"]
        elif key["KeyType"] == "RANGE":
            key_range = key["AttributeName"]

    if key_hash is not None and key_range is not None:
        return key_hash, key_range
    else:
        raise ValueError("Must be a DynamoDB table with a hash key and sort key.")


def convert_float_to_decimal(obj: Any) -> Any:
    if isinstance(obj, List):
        return [convert_float_to_decimal(x) for x in obj]
    elif isinstance(obj, Set):
        return {convert_float_to_decimal(x) for x in obj}
    elif isinstance(obj, Dict):
        return {k: convert_float_to_decimal(v) for k, v in obj.items()}
    elif isinstance(obj, float):
        return Decimal(str(obj))
    else:
        return obj


def convert_decimal_to_int_or_float(obj: Any) -> Any:
    if isinstance(obj, List):
        return [convert_decimal_to_int_or_float(x) for x in obj]
    elif isinstance(obj, Set):
        return {convert_decimal_to_int_or_float(x) for x in obj}
    elif isinstance(obj, Dict):
        return {k: convert_decimal_to_int_or_float(v) for k, v in obj.items()}
    elif isinstance(obj, Decimal):
        return num if (num := int(obj)) == obj else float(obj)
    else:
        return obj


class DynamoDBKVStore(BaseKVStore):
    """DynamoDB Key-Value store.
    Stores key-value pairs in a DynamoDB Table.
    The DynamoDB Table must have both a hash key and a range key,
        and their types must be string.

    You can specify a custom URL for DynamoDB by setting the `DYNAMODB_URL`
    environment variable. This is useful if you're using a local instance of
    DynamoDB for development or testing. If `DYNAMODB_URL` is not set, the
    application will use the default AWS DynamoDB service.

    Args:
        table (Any): DynamoDB Table Service Resource
    """

    def __init__(self, table: Any):
        """Init a DynamoDBKVStore."""
        try:
            from boto3.dynamodb.conditions import Key
        except ImportError:
            raise ImportError(IMPORT_ERROR_MSG)

        self._table = table
        self._boto3_key = Key
        self._key_hash, self._key_range = parse_schema(table)

    @classmethod
    def from_table_name(cls, table_name: str) -> DynamoDBKVStore:
        """Load a DynamoDBKVStore from a DynamoDB table name.

        Args:
            table_name (str): DynamoDB table name
        """
        try:
            import boto3
        except ImportError:
            raise ImportError(IMPORT_ERROR_MSG)

        # Get the DynamoDB URL from environment variable
        dynamodb_url = os.getenv("DYNAMODB_URL")

        # Create a session
        session = boto3.Session()

        # If the DynamoDB URL is set, use it as the endpoint URL
        if dynamodb_url:
            ddb = session.resource("dynamodb", endpoint_url=dynamodb_url)
        else:
            # Otherwise, let boto3 use its default configuration
            ddb = session.resource("dynamodb")
        return cls(table=ddb.Table(table_name))

    def put(self, key: str, val: dict, collection: str = DEFAULT_COLLECTION) -> None:
        """Put a key-value pair into the store.

        Args:
            key (str): key
            val (dict): value
            collection (str): collection name
        """
        item = {k: convert_float_to_decimal(v) for k, v in val.items()}
        item[self._key_hash] = collection
        item[self._key_range] = key
        self._table.put_item(Item=item)

    def get(self, key: str, collection: str = DEFAULT_COLLECTION) -> Optional[dict]:
        """Get a value from the store.

        Args:
            key (str): key
            collection (str): collection name
        """
        resp = self._table.get_item(
            Key={self._key_hash: collection, self._key_range: key}
        )
        if (item := resp.get("Item")) is None:
            return None
        else:
            return {
                k: convert_decimal_to_int_or_float(v)
                for k, v in item.items()
                if k not in {self._key_hash, self._key_range}
            }

    def get_all(self, collection: str = DEFAULT_COLLECTION) -> Dict[str, dict]:
        """Get all values from the store.

        Args:
            collection (str): collection name
        """
        result = {}
        last_evaluated_key = None
        is_first = True
        while last_evaluated_key is not None or is_first:
            if is_first:
                is_first = False
            option = {
                "KeyConditionExpression": self._boto3_key(self._key_hash).eq(collection)
            }
            if last_evaluated_key is not None:
                option["ExclusiveStartKey"] = last_evaluated_key
            resp = self._table.query(**option)
            for item in resp.get("Items", []):
                item.pop(self._key_hash)
                key = item.pop(self._key_range)
                result[key] = {
                    k: convert_decimal_to_int_or_float(v) for k, v in item.items()
                }
            last_evaluated_key = resp.get("LastEvaluatedKey")
        return result

    def delete(self, key: str, collection: str = DEFAULT_COLLECTION) -> bool:
        """Delete a value from the store.

        Args:
            key (str): key
            collection (str): collection name
        """
        resp = self._table.delete_item(
            Key={self._key_hash: collection, self._key_range: key},
            ReturnValues="ALL_OLD",
        )

        if (item := resp.get("Attributes")) is None:
            return False
        else:
            return len(item) > 0



"""OpenAI Finetuning."""

from llama_index.finetuning.types import BaseLLMFinetuneEngine
from llama_index.callbacks import OpenAIFineTuningHandler
from llama_index.finetuning.openai.validate_json import validate_json
from typing import Any, Optional
import os
import openai
from llama_index.llms import OpenAI
import time
import logging

from llama_index.llms.base import LLM

logger = logging.getLogger(__name__)


class OpenAIFinetuneEngine(BaseLLMFinetuneEngine):
    """OpenAI Finetuning Engine."""

    def __init__(
        self,
        base_model: str,
        data_path: str,
        verbose: bool = False,
        start_job_id: Optional[str] = None,
    ) -> None:
        """Init params."""
        self.base_model = base_model
        self.data_path = data_path
        self._verbose = verbose
        self._start_job: Optional[Any] = None
        if start_job_id is not None:
            self._start_job = openai.FineTuningJob.retrieve(start_job_id)

    @classmethod
    def from_finetuning_handler(
        cls,
        finetuning_handler: OpenAIFineTuningHandler,
        base_model: str,
        data_path: str,
        **kwargs: Any,
    ) -> "OpenAIFinetuneEngine":
        """Initialize from finetuning handler.

        Used to finetune an OpenAI model into another
        OpenAI model (e.g. gpt-3.5-turbo on top of GPT-4).

        """
        finetuning_handler.save_finetuning_events(data_path)
        return cls(base_model=base_model, data_path=data_path, **kwargs)

    def finetune(self) -> None:
        """Finetune model."""
        validate_json(self.data_path)

        file_name = os.path.basename(self.data_path)

        # upload file
        with open(self.data_path, "rb") as f:
            output = openai.File.create(
                file=f,
                purpose="fine-tune",
                user_provided_filename=file_name,
            )
        logger.info("File uploaded...")
        if self._verbose:
            print("File uploaded...")

        # launch training
        while True:
            try:
                job_output = openai.FineTuningJob.create(
                    training_file=output["id"], model=self.base_model
                )
                self._start_job = job_output
                break
            except openai.error.InvalidRequestError:
                print("Waiting for file to be ready...")
                time.sleep(60)
        info_str = (
            f"Training job {output['id']} launched. "
            "You will be emailed when it's complete."
        )
        logger.info(info_str)
        if self._verbose:
            print(info_str)

    def get_current_job(self) -> Any:
        """Get current job."""
        # validate that it works
        if not self._start_job:
            raise ValueError("Must call finetune() first")

        # try getting id, make sure that run succeeded
        job_id = self._start_job["id"]
        current_job = openai.FineTuningJob.retrieve(job_id)
        return current_job

    def get_finetuned_model(self, **model_kwargs: Any) -> LLM:
        """Gets finetuned model."""
        current_job = self.get_current_job()

        job_id = current_job["id"]
        status = current_job["status"]
        model_id = current_job["fine_tuned_model"]

        if model_id is None:
            raise ValueError(
                f"Job {job_id} does not have a finetuned model id ready yet."
            )
        if status != "succeeded":
            raise ValueError(f"Job {job_id} has status {status}, cannot get model")

        llm = OpenAI(model=model_id, **model_kwargs)
        return llm



from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, cast

from llama_index.utils import globals_helper
from llama_index.callbacks.base_handler import BaseCallbackHandler
from llama_index.callbacks.schema import CBEventType, EventPayload


@dataclass
class TokenCountingEvent:
    prompt: str
    completion: str
    completion_token_count: int
    prompt_token_count: int
    total_token_count: int = 0
    event_id: str = ""

    def __post_init__(self) -> None:
        self.total_token_count = self.prompt_token_count + self.completion_token_count


def get_llm_token_counts(
    tokenizer: Callable[[str], List], payload: Dict[str, Any], event_id: str = ""
) -> TokenCountingEvent:
    from llama_index.llms import ChatMessage

    if EventPayload.PROMPT in payload:
        prompt = str(payload.get(EventPayload.PROMPT))
        completion = str(payload.get(EventPayload.COMPLETION))

        return TokenCountingEvent(
            event_id=event_id,
            prompt=prompt,
            prompt_token_count=len(tokenizer(prompt)),
            completion=completion,
            completion_token_count=len(tokenizer(completion)),
        )

    elif EventPayload.MESSAGES in payload:
        messages = cast(List[ChatMessage], payload.get(EventPayload.MESSAGES, []))
        messages_str = "\n".join([str(x) for x in messages])
        response = str(payload.get(EventPayload.RESPONSE))

        return TokenCountingEvent(
            event_id=event_id,
            prompt=messages_str,
            prompt_token_count=len(tokenizer(messages_str)),
            completion=response,
            completion_token_count=len(tokenizer(response)),
        )
    else:
        raise ValueError(
            "Invalid payload! Need prompt and completion or messages and response."
        )


class TokenCountingHandler(BaseCallbackHandler):
    """Callback handler for counting tokens in LLM and Embedding events.

    Args:
        tokenizer:
            Tokenizer to use. Defaults to the global tokenizer
            (see llama_index.utils.globals_helper).
        event_starts_to_ignore: List of event types to ignore at the start of a trace.
        event_ends_to_ignore: List of event types to ignore at the end of a trace.
    """

    def __init__(
        self,
        tokenizer: Optional[Callable[[str], List]] = None,
        event_starts_to_ignore: Optional[List[CBEventType]] = None,
        event_ends_to_ignore: Optional[List[CBEventType]] = None,
        verbose: bool = False,
    ) -> None:
        self.llm_token_counts: List[TokenCountingEvent] = []
        self.embedding_token_counts: List[TokenCountingEvent] = []
        self.tokenizer = tokenizer or globals_helper.tokenizer
        self._verbose = verbose

        super().__init__(
            event_starts_to_ignore=event_starts_to_ignore or [],
            event_ends_to_ignore=event_ends_to_ignore or [],
        )

    def start_trace(self, trace_id: Optional[str] = None) -> None:
        return

    def end_trace(
        self,
        trace_id: Optional[str] = None,
        trace_map: Optional[Dict[str, List[str]]] = None,
    ) -> None:
        return

    def on_event_start(
        self,
        event_type: CBEventType,
        payload: Optional[Dict[str, Any]] = None,
        event_id: str = "",
        **kwargs: Any,
    ) -> str:
        return event_id

    def on_event_end(
        self,
        event_type: CBEventType,
        payload: Optional[Dict[str, Any]] = None,
        event_id: str = "",
        **kwargs: Any,
    ) -> None:
        """Count the LLM or Embedding tokens as needed."""
        if (
            event_type == CBEventType.LLM
            and event_type not in self.event_ends_to_ignore
            and payload is not None
        ):
            self.llm_token_counts.append(
                get_llm_token_counts(
                    tokenizer=self.tokenizer,
                    payload=payload,
                    event_id=event_id,
                )
            )

            if self._verbose:
                print(
                    "LLM Prompt Token Usage: "
                    f"{self.llm_token_counts[-1].prompt_token_count}\n"
                    "LLM Completion Token Usage: "
                    f"{self.llm_token_counts[-1].completion_token_count}",
                    flush=True,
                )
        elif (
            event_type == CBEventType.EMBEDDING
            and event_type not in self.event_ends_to_ignore
            and payload is not None
        ):
            total_chunk_tokens = 0
            for chunk in payload.get(EventPayload.CHUNKS, []):
                self.embedding_token_counts.append(
                    TokenCountingEvent(
                        event_id=event_id,
                        prompt=chunk,
                        prompt_token_count=len(self.tokenizer(chunk)),
                        completion="",
                        completion_token_count=0,
                    )
                )
                total_chunk_tokens += self.embedding_token_counts[-1].total_token_count

            if self._verbose:
                print(f"Embedding Token Usage: {total_chunk_tokens}", flush=True)

    @property
    def total_llm_token_count(self) -> int:
        """Get the current total LLM token count."""
        return sum([x.total_token_count for x in self.llm_token_counts])

    @property
    def prompt_llm_token_count(self) -> int:
        """Get the current total LLM prompt token count."""
        return sum([x.prompt_token_count for x in self.llm_token_counts])

    @property
    def completion_llm_token_count(self) -> int:
        """Get the current total LLM completion token count."""
        return sum([x.completion_token_count for x in self.llm_token_counts])

    @property
    def total_embedding_token_count(self) -> int:
        """Get the current total Embedding token count."""
        return sum([x.total_token_count for x in self.embedding_token_counts])

    def reset_counts(self) -> None:
        """Reset the token counts."""
        self.llm_token_counts = []
        self.embedding_token_counts = []



"""Test optimization."""

import pytest
from typing import Any, List
from unittest.mock import patch

from llama_index.embeddings.openai import OpenAIEmbedding
from llama_index.indices.query.schema import QueryBundle
from llama_index.indices.postprocessor.optimizer import SentenceEmbeddingOptimizer
from llama_index.schema import TextNode, NodeWithScore
from llama_index.utils import (
    get_transformer_tokenizer_fn,
)

try:
    from transformers import AutoTokenizer
except ImportError:
    AutoTokenizer = None  # type: ignore


def mock_tokenizer_fn(text: str) -> List[str]:
    """Mock tokenizer function."""
    # split by words
    return text.split(" ")


def mock_tokenizer_fn2(text: str) -> List[str]:
    """Mock tokenizer function."""
    # split by words
    return text.split(",")


def mock_get_text_embedding(text: str) -> List[float]:
    """Mock get text embedding."""
    # assume dimensions are 5
    if text == "hello":
        return [1, 0, 0, 0, 0]
    elif text == "world":
        return [0, 1, 0, 0, 0]
    elif text == "foo":
        return [0, 0, 1, 0, 0]
    elif text == "bar":
        return [0, 0, 0, 1, 0]
    elif text == "abc":
        return [0, 0, 0, 0, 1]
    else:
        raise ValueError("Invalid text for `mock_get_text_embedding`.")


def mock_get_text_embeddings(texts: List[str]) -> List[List[float]]:
    """Mock get text embeddings."""
    return [mock_get_text_embedding(text) for text in texts]


def mock_get_text_embedding_chinese(text: str) -> List[float]:
    """Mock get text embedding."""
    # assume dimensions are 5
    if text == "你":
        return [1, 0, 0, 0, 0]
    elif text == "好":
        return [0, 1, 0, 0, 0]
    elif text == "世":
        return [0, 0, 1, 0, 0]
    elif text == "界":
        return [0, 0, 0, 1, 0]
    elif text == "abc":
        return [0, 0, 0, 0, 1]
    else:
        raise ValueError("Invalid text for `mock_get_text_embedding_chinese`.", text)


def mock_get_text_embeddings_chinese(texts: List[str]) -> List[List[float]]:
    """Mock get text embeddings."""
    return [mock_get_text_embedding_chinese(text) for text in texts]


@patch.object(
    OpenAIEmbedding, "_get_text_embedding", side_effect=mock_get_text_embedding
)
@patch.object(
    OpenAIEmbedding, "_get_text_embeddings", side_effect=mock_get_text_embeddings
)
def test_optimizer(_mock_embeds: Any, _mock_embed: Any) -> None:
    """Test optimizer."""
    optimizer = SentenceEmbeddingOptimizer(
        tokenizer_fn=mock_tokenizer_fn, percentile_cutoff=0.5
    )
    query = QueryBundle(query_str="hello", embedding=[1, 0, 0, 0, 0])
    orig_node = TextNode(text="hello world")
    optimized_node = optimizer.postprocess_nodes(
        [NodeWithScore(node=orig_node)], query
    )[0]
    assert optimized_node.node.get_content() == "hello"

    # test with threshold cutoff
    optimizer = SentenceEmbeddingOptimizer(
        tokenizer_fn=mock_tokenizer_fn, threshold_cutoff=0.3
    )
    query = QueryBundle(query_str="world", embedding=[0, 1, 0, 0, 0])
    orig_node = TextNode(text="hello world")
    optimized_node = optimizer.postprocess_nodes(
        [NodeWithScore(node=orig_node)], query
    )[0]
    assert optimized_node.node.get_content() == "world"

    # test with comma splitter
    optimizer = SentenceEmbeddingOptimizer(
        tokenizer_fn=mock_tokenizer_fn2, threshold_cutoff=0.3
    )
    query = QueryBundle(query_str="foo", embedding=[0, 0, 1, 0, 0])
    orig_node = TextNode(text="hello,world,foo,bar")
    optimized_node = optimizer.postprocess_nodes(
        [NodeWithScore(node=orig_node)], query
    )[0]
    assert optimized_node.node.get_content() == "foo"


@patch.object(
    OpenAIEmbedding, "_get_text_embedding", side_effect=mock_get_text_embeddings_chinese
)
@patch.object(
    OpenAIEmbedding,
    "_get_text_embeddings",
    side_effect=mock_get_text_embeddings_chinese,
)
@pytest.mark.skipif(AutoTokenizer is None, reason="transformers not installed")
def test_optimizer_chinese(_mock_embeds: Any, _mock_embed: Any) -> None:
    """Test optimizer."""
    optimizer = SentenceEmbeddingOptimizer(
        tokenizer_fn=get_transformer_tokenizer_fn("GanymedeNil/text2vec-large-chinese"),
        percentile_cutoff=0.5,
    )
    query = QueryBundle(query_str="你好 世界", embedding=[1, 0, 0, 0, 0])
    orig_node = TextNode(text="你好 世界")
    optimized_node = optimizer.postprocess_nodes(
        [NodeWithScore(node=orig_node)], query
    )[0]
    assert len(optimized_node.node.get_content()) < len(
        TextNode(text="你好 世界").get_content()
    )



from inspect import signature
from typing import Any, Callable, List, Optional, Tuple, Type, Union, cast

from llama_index.bridge.pydantic import BaseModel, create_model, FieldInfo


def create_schema_from_function(
    name: str,
    func: Callable[..., Any],
    additional_fields: Optional[
        List[Union[Tuple[str, Type, Any], Tuple[str, Type]]]
    ] = None,
) -> Type[BaseModel]:
    """Create schema from function."""
    fields = {}
    params = signature(func).parameters
    for param_name in params.keys():
        param_type = params[param_name].annotation
        param_default = params[param_name].default

        if param_type is params[param_name].empty:
            param_type = Any

        if param_default is params[param_name].empty:
            # Required field
            fields[param_name] = (param_type, FieldInfo())
        elif isinstance(param_default, FieldInfo):
            # Field with pydantic.Field as default value
            fields[param_name] = (param_type, param_default)
        else:
            fields[param_name] = (param_type, FieldInfo(default=param_default))

    additional_fields = additional_fields or []
    for field_info in additional_fields:
        if len(field_info) == 3:
            field_info = cast(Tuple[str, Type, Any], field_info)
            field_name, field_type, field_default = field_info
            fields[field_name] = (field_type, FieldInfo(default=field_default))
        elif len(field_info) == 2:
            # Required field has no default value
            field_info = cast(Tuple[str, Type], field_info)
            field_name, field_type = field_info
            fields[field_name] = (field_type, FieldInfo())
        else:
            raise ValueError(
                f"Invalid additional field info: {field_info}. "
                "Must be a tuple of length 2 or 3."
            )

    return create_model(name, **fields)  # type: ignore



"""Test pydantic output parser."""

import pytest

from llama_index.bridge.pydantic import BaseModel

from llama_index.output_parsers.pydantic import PydanticOutputParser


class AttrDict(BaseModel):
    test_attr: str
    foo: int


class TestModel(BaseModel):
    title: str
    attr_dict: AttrDict


def test_pydantic() -> None:
    """Test pydantic output parser."""
    output = """\
    
    Here is the valid JSON:
    {
        "title": "TestModel",
        "attr_dict": {
            "test_attr": "test_attr",
            "foo": 2
        }
    }
    """

    parser = PydanticOutputParser(output_cls=TestModel)
    parsed_output = parser.parse(output)
    assert isinstance(parsed_output, TestModel)
    assert parsed_output.title == "TestModel"
    assert isinstance(parsed_output.attr_dict, AttrDict)
    assert parsed_output.attr_dict.test_attr == "test_attr"
    assert parsed_output.attr_dict.foo == 2

    # TODO: figure out testing conditions
    with pytest.raises(ValueError):
        output = "hello world"
        parsed_output = parser.parse(output)


def test_pydantic_format() -> None:
    """Test pydantic format."""
    query = "hello world"
    parser = PydanticOutputParser(output_cls=AttrDict)
    formatted_query = parser.format(query)
    assert "hello world" in formatted_query



"""Awadb reader."""

from typing import Any, List

import numpy as np

from llama_index.readers.base import BaseReader
from llama_index.schema import Document


class AwadbReader(BaseReader):
    """Awadb reader.

    Retrieves documents through an existing awadb client.
    These documents can then be used in a downstream LlamaIndex data structure.

    Args:
        client (awadb.client): An awadb client.

    """

    def __init__(self, client: Any):
        """Initialize with parameters."""
        import_err_msg = """
            `faiss` package not found. For instructions on
            how to install `faiss` please visit
            https://github.com/facebookresearch/faiss/wiki/Installing-Faiss
        """
        try:
            pass
        except ImportError:
            raise ImportError(import_err_msg)

        self.awadb_client = client

    def load_data(
        self,
        query: np.ndarray,
        k: int = 4,
        separate_documents: bool = True,
    ) -> List[Document]:
        """Load data from Faiss.

        Args:
            query (np.ndarray): A 2D numpy array of query vectors.
            k (int): Number of nearest neighbors to retrieve. Defaults to 4.
            separate_documents (Optional[bool]): Whether to return separate
                documents. Defaults to True.
        Returns:
            List[Document]: A list of documents.

        """
        results = self.awadb_client.Search(
            query,
            k,
            text_in_page_content=None,
            meta_filter=None,
            not_include_fields=None,
        )
        documents = []
        for item_detail in results[0]["ResultItems"]:
            documents.append(Document(text=item_detail["embedding_text"]))

        if not separate_documents:
            # join all documents into one
            text_list = [doc.get_content() for doc in documents]
            text = "\n\n".join(text_list)
            documents = [Document(text=text)]

        return documents



from llama_index.output_parsers.utils import extract_json_str


def test_extract_json_str() -> None:
    input = """\
Here is the valid JSON:
{
    "title": "TestModel",
    "attr_dict": {
        "test_attr": "test_attr",
        "foo": 2
    }
}\
"""
    expected = """\
{
    "title": "TestModel",
    "attr_dict": {
        "test_attr": "test_attr",
        "foo": 2
    }
}\
"""
    assert extract_json_str(input) == expected



"""Test LLM reranker."""

from typing import Any, List
from unittest.mock import patch

from llama_index.indices.postprocessor.llm_rerank import LLMRerank
from llama_index.indices.query.schema import QueryBundle
from llama_index.indices.service_context import ServiceContext
from llama_index.llm_predictor import LLMPredictor
from llama_index.prompts import BasePromptTemplate
from llama_index.schema import BaseNode, NodeWithScore, TextNode


def mock_llmpredictor_predict(
    self: Any, prompt: BasePromptTemplate, **prompt_args: Any
) -> str:
    """Patch llm predictor predict."""
    context_str = prompt_args["context_str"]
    node_strs = context_str.split("\n")
    node_to_choice_and_score = {
        "Test": (True, "1"),
        "Test2": (False, "0"),
        "Test3": (True, "3"),
        "Test4": (False, "0"),
        "Test5": (True, "5"),
        "Test6": (False, "0"),
        "Test7": (True, "7"),
        "Test8": (False, "0"),
    }
    choices_and_scores = []
    for idx, node_str in enumerate(node_strs):
        choice, score = node_to_choice_and_score[node_str]
        if choice:
            choices_and_scores.append((idx + 1, score))

    result_strs = [f"Doc: {str(c)}, Relevance: {s}" for c, s in choices_and_scores]
    return "\n".join(result_strs)


def mock_format_node_batch_fn(nodes: List[BaseNode]) -> str:
    """Mock format node batch fn."""
    return "\n".join([node.get_content() for node in nodes])


@patch.object(
    LLMPredictor,
    "predict",
    mock_llmpredictor_predict,
)
def test_llm_rerank(mock_service_context: ServiceContext) -> None:
    """Test LLM rerank."""
    nodes = [
        TextNode(text="Test"),
        TextNode(text="Test2"),
        TextNode(text="Test3"),
        TextNode(text="Test4"),
        TextNode(text="Test5"),
        TextNode(text="Test6"),
        TextNode(text="Test7"),
        TextNode(text="Test8"),
    ]
    nodes_with_score = [NodeWithScore(node=n) for n in nodes]

    # choice batch size 4 (so two batches)
    # take top-3 across all data
    llm_rerank = LLMRerank(
        format_node_batch_fn=mock_format_node_batch_fn,
        choice_batch_size=4,
        top_n=3,
        service_context=mock_service_context,
    )
    query_str = "What is?"
    result_nodes = llm_rerank.postprocess_nodes(
        nodes_with_score, QueryBundle(query_str)
    )
    assert len(result_nodes) == 3
    assert result_nodes[0].node.get_content() == "Test7"
    assert result_nodes[1].node.get_content() == "Test5"
    assert result_nodes[2].node.get_content() == "Test3"



from typing import Callable, List, Dict
from llama_index.utils import get_cache_dir
from llama_index.schema import Document
from llama_index.indices.base_retriever import BaseRetriever
import os
import tqdm
from shutil import rmtree


class BeirEvaluator:
    """
    Refer to: https://github.com/beir-cellar/beir for a full list of supported datasets
    and a full description of BEIR.
    """

    def __init__(self) -> None:
        try:
            pass
        except ImportError:
            raise ImportError(
                "Please install beir to use this feature: " "`pip install beir`",
            )

    def _download_datasets(self, datasets: List[str] = ["nfcorpus"]) -> Dict[str, str]:
        from beir import util

        cache_dir = get_cache_dir()

        dataset_paths = {}
        for dataset in datasets:
            dataset_full_path = os.path.join(cache_dir, "datasets", "BeIR__" + dataset)
            if not os.path.exists(dataset_full_path):
                url = f"""https://public.ukp.informatik.tu-darmstadt.de/thakur\
/BEIR/datasets/{dataset}.zip"""
                try:
                    util.download_and_unzip(url, dataset_full_path)
                except Exception as e:
                    print(
                        "Dataset:", dataset, "not found at:", url, "Removing cached dir"
                    )
                    rmtree(dataset_full_path)
                    raise ValueError(f"invalid BEIR dataset: {dataset}") from e

            print("Dataset:", dataset, "downloaded at:", dataset_full_path)
            dataset_paths[dataset] = os.path.join(dataset_full_path, dataset)
        return dataset_paths

    def run(
        self,
        create_retriever: Callable[[List[Document]], BaseRetriever],
        datasets: List[str] = ["nfcorpus"],
        metrics_k_values: List[int] = [3, 10],
    ) -> None:
        from beir.datasets.data_loader import GenericDataLoader
        from beir.retrieval.evaluation import EvaluateRetrieval

        dataset_paths = self._download_datasets(datasets)
        for dataset in datasets:
            dataset_path = dataset_paths[dataset]
            print("Evaluating on dataset:", dataset)
            print("-------------------------------------")

            corpus, queries, qrels = GenericDataLoader(data_folder=dataset_path).load(
                split="test"
            )

            documents = []
            for id, val in corpus.items():
                doc = Document(
                    text=val["text"], metadata={"title": val["title"], "doc_id": id}
                )
                documents.append(doc)

            retriever = create_retriever(documents)

            print("Retriever created for: ", dataset)

            print("Evaluating retriever on questions against qrels")

            results = {}
            for key, query in tqdm.tqdm(queries.items()):
                nodes_with_score = retriever.retrieve(query)
                results[key] = {
                    node.node.metadata["doc_id"]: node.score
                    for node in nodes_with_score
                }

            ndcg, map_, recall, precision = EvaluateRetrieval.evaluate(
                qrels, results, metrics_k_values
            )
            print("Results for:", dataset)
            for k in metrics_k_values:
                print(
                    {
                        f"NDCG@{k}": ndcg[f"NDCG@{k}"],
                        f"MAP@{k}": map_[f"MAP@{k}"],
                        f"Recall@{k}": recall[f"Recall@{k}"],
                        f"precision@{k}": precision[f"P@{k}"],
                    }
                )
            print("-------------------------------------")



from __future__ import annotations

from typing import TYPE_CHECKING, Any, List, Tuple
from uuid import uuid4

from llama_index.llm_predictor.vellum.types import (
    VellumCompiledPrompt,
    VellumRegisteredPrompt,
)
from llama_index.llm_predictor.vellum.utils import convert_to_kebab_case
from llama_index.prompts import BasePromptTemplate
from llama_index.prompts.base import PromptTemplate

if TYPE_CHECKING:
    import vellum  # noqa: F401


class VellumPromptRegistry:
    """Registers and retrieves prompts with Vellum.

    LlamaIndex Prompts can be registered within Vellum, at which point Vellum becomes
    the source of truth for the prompt. From there, Vellum can be used for prompt/model
    experimentation, request monitoring, and more.
    """

    def __init__(self, vellum_api_key: str) -> None:
        import_err_msg = (
            "`vellum` package not found, please run `pip install vellum-ai`"
        )
        try:
            from vellum.client import Vellum  # noqa: F401
        except ImportError:
            raise ImportError(import_err_msg)

        self._vellum_client = Vellum(api_key=vellum_api_key)

    def from_prompt(self, initial_prompt: BasePromptTemplate) -> VellumRegisteredPrompt:
        """Accepts a LlamaIndex prompt and retrieves a corresponding registered prompt
        from Vellum.

        If the LlamaIndex prompt hasn't yet been registered, it'll be registered
        automatically, after which point Vellum becomes the source-of-truth for the
        prompt's definition.

        In this way, the LlamaIndex prompt is treated as the initial value for the newly
        registered prompt in Vellum.

        You can reference a previously registered prompt by providing either
        `vellum_deployment_id` or `vellum_deployment_name` as key/value pairs within
        `BasePromptTemplate.metadata`.
        """

        from vellum.core import ApiError

        deployment_id = initial_prompt.metadata.get("vellum_deployment_id")
        deployment_name = initial_prompt.metadata.get(
            "vellum_deployment_name"
        ) or self._generate_default_name(initial_prompt)

        registered_prompt: VellumRegisteredPrompt
        try:
            deployment = self._vellum_client.deployments.retrieve(
                deployment_id or deployment_name
            )
        except ApiError as e:
            if e.status_code == 404:
                registered_prompt = self._register_prompt(initial_prompt)
            else:
                raise
        else:
            registered_prompt = self._get_registered_prompt(deployment)

        return registered_prompt

    def get_compiled_prompt(
        self, registered_prompt: VellumRegisteredPrompt, input_values: dict[str, Any]
    ) -> VellumCompiledPrompt:
        """Retrieves the fully-compiled prompt from Vellum, after all variable
        substitutions, templating, etc.
        """

        result = self._vellum_client.model_versions.model_version_compile_prompt(
            registered_prompt.model_version_id, input_values=input_values
        )
        return VellumCompiledPrompt(
            text=result.prompt.text, num_tokens=result.prompt.num_tokens
        )

    def _get_registered_prompt(
        self, deployment: vellum.DeploymentRead
    ) -> VellumRegisteredPrompt:
        """Retrieves a prompt from Vellum, keying off of the deployment's id/name."""

        # Assume that the deployment backing a registered prompt will always have a
        # single model version. Note that this may not be true in the future once
        # deployment-level A/B testing is supported and someone configures an A/B test.
        model_version_id = deployment.active_model_version_ids[0]
        model_version = self._vellum_client.model_versions.retrieve(model_version_id)

        sandbox_snapshot_info = model_version.build_config.sandbox_snapshot
        sandbox_snapshot_id = (
            sandbox_snapshot_info.id if sandbox_snapshot_info else None
        )
        prompt_id = sandbox_snapshot_info.prompt_id if sandbox_snapshot_info else None
        sandbox_id = sandbox_snapshot_info.sandbox_id if sandbox_snapshot_info else None

        return VellumRegisteredPrompt(
            deployment_id=deployment.id,
            deployment_name=deployment.name,
            model_version_id=model_version.id,
            sandbox_id=sandbox_id,
            sandbox_snapshot_id=sandbox_snapshot_id,
            prompt_id=prompt_id,
        )

    def _register_prompt(self, prompt: BasePromptTemplate) -> VellumRegisteredPrompt:
        """Registers a prompt with Vellum.

        By registering a prompt, Vellum will:
        1) Create a Sandbox for the prompt so that you can experiment with the
              prompt, LLM provider, model, and parameters via Vellum's UI.
        2) Deployment for the prompt so that you can monitor requests and
            update the prompt, LLM provider, model, and parameters via Vellum's UI
            without requiring code changes.
        """

        # Label represents a human-friendly name that'll be used for all created
        # entities within Vellum. If not provided, a default will be generated.
        label = prompt.metadata.get(
            "vellum_deployment_label"
        ) or self._generate_default_label(prompt)

        # Name represents a kebab-cased unique identifier that'll be used for all
        # created entities within Vellum. If not provided, a default will be generated.
        name = prompt.metadata.get(
            "vellum_deployment_name"
        ) or self._generate_default_name(prompt)

        # Note: For now, the initial provider, model, and parameters used to register
        # the prompt are hard-coded. You can then update any of these from within
        # Vellum's UI. As a future improvement, we could allow these to be specified
        # upfront.
        provider, model, params = self._get_default_llm_meta()
        prompt_info = self._construct_prompt_info(prompt, for_chat_model=True)

        resp = self._vellum_client.registered_prompts.register_prompt(
            label=label,
            name=name,
            prompt=prompt_info,
            provider=provider,
            model=model,
            parameters=params,
            meta={
                "source": "llamaindex",
                "prompt_type": prompt.metadata["prompt_type"],
            },
        )

        return VellumRegisteredPrompt(
            deployment_id=resp.deployment.id,
            deployment_name=resp.deployment.name,
            model_version_id=resp.model_version.id,
            sandbox_id=resp.sandbox.id,
            sandbox_snapshot_id=resp.sandbox_snapshot.id,
            prompt_id=resp.prompt.id,
        )

    def _generate_default_label(self, prompt: BasePromptTemplate) -> str:
        prompt_type = prompt.metadata["prompt_type"]
        return f"LlamaIndex Demo: {prompt_type}'"

    def _generate_default_name(self, prompt: BasePromptTemplate) -> str:
        default_label = self._generate_default_label(prompt)
        return convert_to_kebab_case(default_label)

    def _construct_prompt_info(
        self, prompt: BasePromptTemplate, for_chat_model: bool = True
    ) -> vellum.RegisterPromptPromptInfoRequest:
        """Converts a LlamaIndex prompt into Vellum's prompt representation."""

        import vellum

        assert isinstance(prompt, PromptTemplate)
        prompt_template = prompt.template
        for input_variable in prompt.template_vars:
            prompt_template = prompt_template.replace(
                input_variable, f"{{ {input_variable} }}"
            )

        block: vellum.PromptTemplateBlockRequest
        jinja_block = vellum.PromptTemplateBlockRequest(
            id=str(uuid4()),
            block_type=vellum.BlockTypeEnum.JINJA,
            properties=vellum.PromptTemplateBlockPropertiesRequest(
                template=self._prepare_prompt_jinja_template(
                    prompt.template,
                    prompt.template_vars,
                ),
            ),
        )
        if for_chat_model:
            block = vellum.PromptTemplateBlockRequest(
                id=str(uuid4()),
                block_type=vellum.BlockTypeEnum.CHAT_MESSAGE,
                properties=vellum.PromptTemplateBlockPropertiesRequest(
                    chat_role=vellum.ChatMessageRole.SYSTEM,
                    blocks=[jinja_block],
                ),
            )
        else:
            block = jinja_block

        return vellum.RegisterPromptPromptInfoRequest(
            prompt_syntax_version=2,
            prompt_block_data=vellum.PromptTemplateBlockDataRequest(
                version=1,
                blocks=[block],
            ),
            input_variables=[{"key": input_var} for input_var in prompt.template_vars],
        )

    def _prepare_prompt_jinja_template(
        self, original_template: str, input_variables: List[str]
    ) -> str:
        """Converts a prompt template into a Jinja template."""

        prompt_template = original_template
        for input_variable in input_variables:
            prompt_template = prompt_template.replace(
                ("{" + input_variable + "}"), ("{{ " + input_variable + " }}")
            )

        return prompt_template

    def _get_default_llm_meta(
        self,
    ) -> Tuple[vellum.ProviderEnum, str, vellum.RegisterPromptModelParametersRequest]:
        import vellum

        return (
            vellum.ProviderEnum.OPENAI,
            "gpt-3.5-turbo",
            vellum.RegisterPromptModelParametersRequest(
                temperature=0.0,
                max_tokens=256,
                stop=[],
                top_p=1.0,
                top_k=0.0,
                frequency_penalty=0.0,
                presence_penalty=0.0,
                logit_bias=None,
            ),
        )



"""Test text splitter."""
import os

from llama_index.text_splitter import CodeSplitter


def test_python_code_splitter() -> None:
    """Test case for code splitting using python"""

    if "CI" in os.environ:
        return

    code_splitter = CodeSplitter(
        language="python", chunk_lines=4, chunk_lines_overlap=1, max_chars=30
    )

    text = """\
def foo():
    print("bar")

def baz():
    print("bbq")"""

    chunks = code_splitter.split_text(text)
    assert chunks[0].startswith("def foo():")
    assert chunks[1].startswith("def baz():")


def test_typescript_code_splitter() -> None:
    """Test case for code splitting using typescript"""

    if "CI" in os.environ:
        return

    code_splitter = CodeSplitter(
        language="typescript", chunk_lines=4, chunk_lines_overlap=1, max_chars=50
    )

    text = """\
function foo() {
    console.log("bar");
}

function baz() {
    console.log("bbq");
}"""

    chunks = code_splitter.split_text(text)
    assert chunks[0].startswith("function foo()")
    assert chunks[1].startswith("function baz()")


def test_html_code_splitter() -> None:
    """Test case for code splitting using typescript"""

    if "CI" in os.environ:
        return

    code_splitter = CodeSplitter(
        language="html", chunk_lines=4, chunk_lines_overlap=1, max_chars=50
    )

    text = """\
<!DOCTYPE html>
<html>
<head>
    <title>My Example Page</title>
</head>
<body>
    <h1>Welcome to My Example Page</h1>
    <p>This is a basic HTML page example.</p>
    <ul>
        <li>Item 1</li>
        <li>Item 2</li>
        <li>Item 3</li>
    </ul>
    <img src="https://example.com/image.jpg" alt="Example Image">
</body>
</html>"""

    chunks = code_splitter.split_text(text)
    assert chunks[0].startswith("<!DOCTYPE html>")
    assert chunks[1].startswith("<html>")
    assert chunks[2].startswith("<head>")


def test_tsx_code_splitter() -> None:
    """Test case for code splitting using typescript"""

    if "CI" in os.environ:
        return

    code_splitter = CodeSplitter(
        language="typescript", chunk_lines=4, chunk_lines_overlap=1, max_chars=50
    )

    text = """\
import React from 'react';

interface Person {
  name: string;
  age: number;
}

const ExampleComponent: React.FC = () => {
  const person: Person = {
    name: 'John Doe',
    age: 30,
  };

  return (
    <div>
      <h1>Hello, {person.name}!</h1>
      <p>You are {person.age} years old.</p>
    </div>
  );
};

export default ExampleComponent;"""

    chunks = code_splitter.split_text(text)
    assert chunks[0].startswith("import React from 'react';")
    assert chunks[1].startswith("interface Person")


def test_cpp_code_splitter() -> None:
    """Test case for code splitting using typescript"""

    if "CI" in os.environ:
        return

    code_splitter = CodeSplitter(
        language="cpp", chunk_lines=4, chunk_lines_overlap=1, max_chars=50
    )

    text = """\
#include <iostream>

int main() {
    std::cout << "Hello, World!" << std::endl;
    return 0;
}"""

    chunks = code_splitter.split_text(text)
    assert chunks[0].startswith("#include <iostream>")
    assert chunks[1].startswith("int main()")
    assert chunks[2].startswith("{\n    std::cout")



from typing import Sequence

import pytest

from llama_index.llms.base import ChatMessage, MessageRole
from llama_index.llms.llama_utils import (
    B_INST,
    B_SYS,
    BOS,
    DEFAULT_SYSTEM_PROMPT,
    E_INST,
    E_SYS,
    EOS,
    completion_to_prompt,
    messages_to_prompt,
)


@pytest.fixture
def chat_messages_first_chat() -> Sequence[ChatMessage]:
    # example first chat with system message
    return [
        ChatMessage(role=MessageRole.SYSTEM, content="some system message"),
        ChatMessage(role=MessageRole.USER, content="test question"),
    ]


@pytest.fixture
def chat_messages_first_chat_no_system(
    chat_messages_first_chat: Sequence[ChatMessage],
) -> Sequence[ChatMessage]:
    # example first chat without system message
    return chat_messages_first_chat[1:]


@pytest.fixture
def chat_messages_second_chat() -> Sequence[ChatMessage]:
    # example second chat with system message
    return [
        ChatMessage(role=MessageRole.SYSTEM, content="some system message"),
        ChatMessage(role=MessageRole.USER, content="test question 1"),
        ChatMessage(role=MessageRole.ASSISTANT, content="some assistant reply"),
        ChatMessage(role=MessageRole.USER, content="test question 2"),
    ]


@pytest.fixture
def chat_messages_second_chat_no_system(
    chat_messages_second_chat: Sequence[ChatMessage],
) -> Sequence[ChatMessage]:
    # example second chat without system message
    return chat_messages_second_chat[1:]


@pytest.fixture
def chat_messages_third_chat() -> Sequence[ChatMessage]:
    # example third chat with system message
    return [
        ChatMessage(role=MessageRole.SYSTEM, content="some system message"),
        ChatMessage(role=MessageRole.USER, content="test question 1"),
        ChatMessage(role=MessageRole.ASSISTANT, content="some assistant reply 1"),
        ChatMessage(role=MessageRole.USER, content="test question 2"),
        ChatMessage(role=MessageRole.ASSISTANT, content="some assistant reply 2"),
        ChatMessage(role=MessageRole.USER, content="test question 3"),
    ]


@pytest.fixture
def chat_messages_third_chat_no_system(
    chat_messages_third_chat: Sequence[ChatMessage],
) -> Sequence[ChatMessage]:
    # example third chat without system message
    return chat_messages_third_chat[1:]


@pytest.fixture
def chat_messages_assistant_first() -> Sequence[ChatMessage]:
    # assistant message first in chat (after system)
    # should raise error as we expect the first message after any system
    # message to be a user message
    return [
        ChatMessage(role=MessageRole.SYSTEM, content="some system message"),
        ChatMessage(role=MessageRole.ASSISTANT, content="some assistant reply"),
        ChatMessage(role=MessageRole.USER, content="test question"),
    ]


@pytest.fixture
def chat_messages_user_twice() -> Sequence[ChatMessage]:
    # user message twice in a row (after system)
    # should raise error as we expect an assistant message
    # to follow a user message
    return [
        ChatMessage(role=MessageRole.SYSTEM, content="some system message"),
        ChatMessage(role=MessageRole.USER, content="test question 1"),
        ChatMessage(role=MessageRole.USER, content="test question 2"),
    ]


def test_first_chat(chat_messages_first_chat: Sequence[ChatMessage]) -> None:
    # test first chat prompt creation with system prompt
    prompt = messages_to_prompt(chat_messages_first_chat)
    assert prompt == (
        f"{BOS} {B_INST} {B_SYS} some system message {E_SYS} test question {E_INST}"
    )


def test_first_chat_default(
    chat_messages_first_chat_no_system: Sequence[ChatMessage],
) -> None:
    # test first chat prompt creation without system prompt and use default
    prompt = messages_to_prompt(chat_messages_first_chat_no_system)
    assert prompt == (
        f"{BOS} {B_INST} {B_SYS} {DEFAULT_SYSTEM_PROMPT.strip()} {E_SYS} "
        f"test question {E_INST}"
    )


def test_second_chat(chat_messages_second_chat: Sequence[ChatMessage]) -> None:
    # test second chat prompt creation with system prompt
    prompt = messages_to_prompt(chat_messages_second_chat)
    assert prompt == (
        f"{BOS} {B_INST} {B_SYS} some system message {E_SYS} "
        f"test question 1 {E_INST} some assistant reply {EOS}"
        f"{BOS} {B_INST} test question 2 {E_INST}"
    )


def test_second_chat_default(
    chat_messages_second_chat_no_system: Sequence[ChatMessage],
) -> None:
    # test second chat prompt creation without system prompt and use default
    prompt = messages_to_prompt(chat_messages_second_chat_no_system)
    assert prompt == (
        f"{BOS} {B_INST} {B_SYS} {DEFAULT_SYSTEM_PROMPT.strip()} {E_SYS} "
        f"test question 1 {E_INST} some assistant reply {EOS}"
        f"{BOS} {B_INST} test question 2 {E_INST}"
    )


def test_third_chat(chat_messages_third_chat: Sequence[ChatMessage]) -> None:
    # test third chat prompt creation with system prompt
    prompt = messages_to_prompt(chat_messages_third_chat)
    assert prompt == (
        f"{BOS} {B_INST} {B_SYS} some system message {E_SYS} "
        f"test question 1 {E_INST} some assistant reply 1 {EOS}"
        f"{BOS} {B_INST} test question 2 {E_INST} some assistant reply 2 {EOS}"
        f"{BOS} {B_INST} test question 3 {E_INST}"
    )


def test_third_chat_default(
    chat_messages_third_chat_no_system: Sequence[ChatMessage],
) -> None:
    # test third chat prompt creation without system prompt and use default
    prompt = messages_to_prompt(chat_messages_third_chat_no_system)
    assert prompt == (
        f"{BOS} {B_INST} {B_SYS} {DEFAULT_SYSTEM_PROMPT.strip()} {E_SYS} "
        f"test question 1 {E_INST} some assistant reply 1 {EOS}"
        f"{BOS} {B_INST} test question 2 {E_INST} some assistant reply 2 {EOS}"
        f"{BOS} {B_INST} test question 3 {E_INST}"
    )


def test_error_assistant_first(
    chat_messages_assistant_first: Sequence[ChatMessage],
) -> None:
    # should have error if assistant message occurs first
    with pytest.raises(AssertionError):
        messages_to_prompt(chat_messages_assistant_first)


def test_error_user_twice(chat_messages_user_twice: Sequence[ChatMessage]) -> None:
    # should have error if second message is user
    # (or have user twice in a row)
    with pytest.raises(AssertionError):
        messages_to_prompt(chat_messages_user_twice)


def test_completion_to_prompt() -> None:
    # test prompt creation from completion with system prompt
    completion = "test completion"
    system_prompt = "test system prompt"
    prompt = completion_to_prompt(completion, system_prompt=system_prompt)
    assert prompt == (
        f"{BOS} {B_INST} {B_SYS} {system_prompt} {E_SYS} {completion} {E_INST}"
    )


def test_completion_to_prompt_default() -> None:
    # test prompt creation from completion without system prompt and use default
    completion = "test completion"
    prompt = completion_to_prompt(completion)
    assert prompt == (
        f"{BOS} {B_INST} {B_SYS} {DEFAULT_SYSTEM_PROMPT.strip()} {E_SYS} "
        f"{completion} {E_INST}"
    )



"""LLM reranker."""
from typing import Callable, List, Optional

from llama_index.bridge.pydantic import Field, PrivateAttr

from llama_index.indices.postprocessor.types import BaseNodePostprocessor
from llama_index.indices.query.schema import QueryBundle
from llama_index.indices.service_context import ServiceContext
from llama_index.indices.utils import (
    default_format_node_batch_fn,
    default_parse_choice_select_answer_fn,
)
from llama_index.prompts import BasePromptTemplate
from llama_index.prompts.default_prompts import DEFAULT_CHOICE_SELECT_PROMPT
from llama_index.schema import NodeWithScore


class LLMRerank(BaseNodePostprocessor):
    """LLM-based reranker."""

    top_n: int = Field(description="Top N nodes to return.")
    choice_select_prompt: BasePromptTemplate = Field(
        description="Choice select prompt."
    )
    choice_batch_size: int = Field(description="Batch size for choice select.")
    service_context: ServiceContext = Field(
        description="Service context.", exclude=True
    )

    _format_node_batch_fn: Callable = PrivateAttr()
    _parse_choice_select_answer_fn: Callable = PrivateAttr()

    def __init__(
        self,
        choice_select_prompt: Optional[BasePromptTemplate] = None,
        choice_batch_size: int = 10,
        format_node_batch_fn: Optional[Callable] = None,
        parse_choice_select_answer_fn: Optional[Callable] = None,
        service_context: Optional[ServiceContext] = None,
        top_n: int = 10,
    ) -> None:
        choice_select_prompt = choice_select_prompt or DEFAULT_CHOICE_SELECT_PROMPT
        service_context = service_context or ServiceContext.from_defaults()

        self._format_node_batch_fn = (
            format_node_batch_fn or default_format_node_batch_fn
        )
        self._parse_choice_select_answer_fn = (
            parse_choice_select_answer_fn or default_parse_choice_select_answer_fn
        )

        super().__init__(
            choice_select_prompt=choice_select_prompt,
            choice_batch_size=choice_batch_size,
            service_context=service_context,
            top_n=top_n,
        )

    @classmethod
    def class_name(cls) -> str:
        return "LLMRerank"

    def postprocess_nodes(
        self,
        nodes: List[NodeWithScore],
        query_bundle: Optional[QueryBundle] = None,
    ) -> List[NodeWithScore]:
        if query_bundle is None:
            raise ValueError("Query bundle must be provided.")
        initial_results: List[NodeWithScore] = []
        for idx in range(0, len(nodes), self.choice_batch_size):
            nodes_batch = [
                node.node for node in nodes[idx : idx + self.choice_batch_size]
            ]

            query_str = query_bundle.query_str
            fmt_batch_str = self._format_node_batch_fn(nodes_batch)
            # call each batch independently
            raw_response = self.service_context.llm_predictor.predict(
                self.choice_select_prompt,
                context_str=fmt_batch_str,
                query_str=query_str,
            )

            raw_choices, relevances = self._parse_choice_select_answer_fn(
                raw_response, len(nodes_batch)
            )
            choice_idxs = [int(choice) - 1 for choice in raw_choices]
            choice_nodes = [nodes_batch[idx] for idx in choice_idxs]
            relevances = relevances or [1.0 for _ in choice_nodes]
            initial_results.extend(
                [
                    NodeWithScore(node=node, score=relevance)
                    for node, relevance in zip(choice_nodes, relevances)
                ]
            )

        results = sorted(initial_results, key=lambda x: x.score or 0.0, reverse=True)[
            : self.top_n
        ]
        return results



from typing import Any, Dict, Type

from llama_index.readers.base import BasePydanticReader
from llama_index.readers.discord_reader import DiscordReader
from llama_index.readers.elasticsearch import ElasticsearchReader
from llama_index.readers.google_readers.gdocs import GoogleDocsReader
from llama_index.readers.google_readers.gsheets import GoogleSheetsReader
from llama_index.readers.notion import NotionPageReader
from llama_index.readers.slack import SlackReader
from llama_index.readers.string_iterable import StringIterableReader
from llama_index.readers.twitter import TwitterTweetReader
from llama_index.readers.web import (
    SimpleWebPageReader,
    TrafilaturaWebReader,
    RssReader,
    BeautifulSoupWebReader,
)
from llama_index.readers.wikipedia import WikipediaReader
from llama_index.readers.youtube_transcript import YoutubeTranscriptReader

ALL_READERS: Dict[str, Type[BasePydanticReader]] = {
    DiscordReader.class_name(): DiscordReader,
    ElasticsearchReader.class_name(): ElasticsearchReader,
    GoogleDocsReader.class_name(): GoogleDocsReader,
    GoogleSheetsReader.class_name(): GoogleSheetsReader,
    NotionPageReader.class_name(): NotionPageReader,
    SlackReader.class_name(): SlackReader,
    StringIterableReader.class_name(): StringIterableReader,
    TwitterTweetReader.class_name(): TwitterTweetReader,
    SimpleWebPageReader.class_name(): SimpleWebPageReader,
    TrafilaturaWebReader.class_name(): TrafilaturaWebReader,
    RssReader.class_name(): RssReader,
    BeautifulSoupWebReader.class_name(): BeautifulSoupWebReader,
    WikipediaReader.class_name(): WikipediaReader,
    YoutubeTranscriptReader.class_name(): YoutubeTranscriptReader,
}


def load_reader(data: Dict[str, Any]) -> BasePydanticReader:
    class_name = data.get("class_name", None)
    if class_name is None:
        raise ValueError("Must specify `class_name` in reader data.")

    if class_name not in ALL_READERS:
        raise ValueError(f"Reader class name {class_name} not found.")

    # remove static attribute
    data.pop("is_remote", None)

    reader = ALL_READERS[class_name].from_dict(data)

    return reader



"""Init file of LlamaIndex."""
from pathlib import Path

with open(Path(__file__).absolute().parents[0] / "VERSION") as _f:
    __version__ = _f.read().strip()


import logging
from logging import NullHandler
from typing import Optional

from llama_index.data_structs.struct_type import IndexStructType

# embeddings
from llama_index.embeddings.langchain import LangchainEmbedding
from llama_index.embeddings.openai import OpenAIEmbedding


# indices
from llama_index.indices.keyword_table import (
    KeywordTableIndex,
    RAKEKeywordTableIndex,
    SimpleKeywordTableIndex,
    GPTKeywordTableIndex,
    GPTRAKEKeywordTableIndex,
    GPTSimpleKeywordTableIndex,
)
from llama_index.indices.knowledge_graph import (
    KnowledgeGraphIndex,
    GPTKnowledgeGraphIndex,
)
from llama_index.indices.list import ListIndex, GPTListIndex, SummaryIndex
from llama_index.indices.tree import TreeIndex, GPTTreeIndex
from llama_index.indices.vector_store import VectorStoreIndex, GPTVectorStoreIndex
from llama_index.indices.document_summary import (
    DocumentSummaryIndex,
    GPTDocumentSummaryIndex,
)
from llama_index.indices.empty import EmptyIndex, GPTEmptyIndex
from llama_index.indices.struct_store.pandas import PandasIndex, GPTPandasIndex
from llama_index.indices.struct_store.sql import (
    SQLStructStoreIndex,
    GPTSQLStructStoreIndex,
)

# structured
from llama_index.indices.common.struct_store.base import SQLDocumentContextBuilder

# for composability
from llama_index.indices.composability.graph import ComposableGraph

# loading
from llama_index.indices.loading import (
    load_graph_from_storage,
    load_index_from_storage,
    load_indices_from_storage,
)

# prompt helper
from llama_index.indices.prompt_helper import PromptHelper

# Response Synthesizer
from llama_index.response_synthesizers.factory import get_response_synthesizer

# QueryBundle
from llama_index.indices.query.schema import QueryBundle

from llama_index.indices.service_context import (
    ServiceContext,
    set_global_service_context,
)

# langchain helper
from llama_index.llm_predictor import LLMPredictor
from llama_index.langchain_helpers.memory_wrapper import GPTIndexMemory
from llama_index.langchain_helpers.sql_wrapper import SQLDatabase

# prompts
from llama_index.prompts import (
    BasePromptTemplate,
    PromptTemplate,
    ChatPromptTemplate,
    SelectorPromptTemplate,
    # backwards compatibility
    Prompt,
)
from llama_index.prompts.prompts import (
    KeywordExtractPrompt,
    QueryKeywordExtractPrompt,
    QuestionAnswerPrompt,
    RefinePrompt,
    SummaryPrompt,
    TreeInsertPrompt,
    TreeSelectMultiplePrompt,
    TreeSelectPrompt,
)

# readers
from llama_index.schema import Document
from llama_index.readers import (
    BeautifulSoupWebReader,
    ChromaReader,
    DeepLakeReader,
    DiscordReader,
    FaissReader,
    GithubRepositoryReader,
    GoogleDocsReader,
    JSONReader,
    MboxReader,
    MilvusReader,
    NotionPageReader,
    ObsidianReader,
    PineconeReader,
    PsychicReader,
    QdrantReader,
    RssReader,
    SimpleDirectoryReader,
    SimpleMongoReader,
    SimpleWebPageReader,
    SlackReader,
    StringIterableReader,
    TrafilaturaWebReader,
    TwitterTweetReader,
    WeaviateReader,
    WikipediaReader,
)
from llama_index.readers.download import download_loader

# response
from llama_index.response.schema import Response

# storage
from llama_index.storage.storage_context import StorageContext

# token predictor
from llama_index.llm_predictor.mock import MockLLMPredictor
from llama_index.token_counter.mock_embed_model import MockEmbedding

# vellum
from llama_index.llm_predictor.vellum import VellumPredictor, VellumPromptRegistry

# import global eval handler
from llama_index.callbacks.global_handlers import set_global_handler

# best practices for library logging:
# https://docs.python.org/3/howto/logging.html#configuring-logging-for-a-library
logging.getLogger(__name__).addHandler(NullHandler())

__all__ = [
    "StorageContext",
    "ServiceContext",
    "ComposableGraph",
    # indices
    "SummaryIndex",
    "VectorStoreIndex",
    "SimpleKeywordTableIndex",
    "KeywordTableIndex",
    "RAKEKeywordTableIndex",
    "TreeIndex",
    "SQLStructStoreIndex",
    "PandasIndex",
    "EmptyIndex",
    "DocumentSummaryIndex",
    "KnowledgeGraphIndex",
    # indices - legacy names
    "GPTKeywordTableIndex",
    "GPTKnowledgeGraphIndex",
    "GPTSimpleKeywordTableIndex",
    "GPTRAKEKeywordTableIndex",
    "GPTListIndex",
    "ListIndex",
    "GPTEmptyIndex",
    "GPTTreeIndex",
    "GPTVectorStoreIndex",
    "GPTPandasIndex",
    "GPTSQLStructStoreIndex",
    "GPTDocumentSummaryIndex",
    "Prompt",
    "PromptTemplate",
    "BasePromptTemplate",
    "ChatPromptTemplate",
    "SelectorPromptTemplate",
    "LangchainEmbedding",
    "OpenAIEmbedding",
    "SummaryPrompt",
    "TreeInsertPrompt",
    "TreeSelectPrompt",
    "TreeSelectMultiplePrompt",
    "RefinePrompt",
    "QuestionAnswerPrompt",
    "KeywordExtractPrompt",
    "QueryKeywordExtractPrompt",
    "Response",
    "WikipediaReader",
    "ObsidianReader",
    "Document",
    "SimpleDirectoryReader",
    "JSONReader",
    "SimpleMongoReader",
    "NotionPageReader",
    "GoogleDocsReader",
    "MboxReader",
    "SlackReader",
    "StringIterableReader",
    "WeaviateReader",
    "FaissReader",
    "ChromaReader",
    "DeepLakeReader",
    "PineconeReader",
    "PsychicReader",
    "QdrantReader",
    "MilvusReader",
    "DiscordReader",
    "SimpleWebPageReader",
    "RssReader",
    "BeautifulSoupWebReader",
    "TrafilaturaWebReader",
    "LLMPredictor",
    "MockLLMPredictor",
    "VellumPredictor",
    "VellumPromptRegistry",
    "MockEmbedding",
    "SQLDatabase",
    "GPTIndexMemory",
    "SQLDocumentContextBuilder",
    "SQLContextBuilder",
    "PromptHelper",
    "IndexStructType",
    "TwitterTweetReader",
    "download_loader",
    "GithubRepositoryReader",
    "load_graph_from_storage",
    "load_index_from_storage",
    "load_indices_from_storage",
    "QueryBundle",
    "get_response_synthesizer",
    "set_global_service_context",
    "set_global_handler",
]

# eval global toggle
from llama_index.callbacks.base_handler import BaseCallbackHandler  # noqa: E402

global_handler: Optional[BaseCallbackHandler] = None

# NOTE: keep for backwards compatibility
SQLContextBuilder = SQLDocumentContextBuilder

# global service context for ServiceContext.from_defaults()
global_service_context: Optional[ServiceContext] = None



from llama_index.prompts.utils import get_template_vars


def test_get_template_vars() -> None:
    template = "hello {text} {foo}"
    template_vars = get_template_vars(template)
    assert template_vars == ["text", "foo"]



from pathlib import Path
import tempfile
import os
import pytest

from llama_index.readers.file.html_reader import HTMLTagReader


@pytest.fixture
def html_str() -> str:
    return """
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>HTML Sections Example</title>
</head>
<body>
    <header>
        <h1>Welcome to My Website</h1>
    </header>

    <nav>
        <ul>
            <li><a href="#">Home</a></li>
            <li><a href="#">About</a></li>
            <li><a href="#">Services</a></li>
            <li><a href="#">Contact</a></li>
        </ul>
    </nav>

    <section id="about">
        <h2>About Us</h2>
        <p>Lorem ipsum dolor sit amet, consectetur adipiscing elit.</p>
    </section>

    <section id="services">
        <h2>Our Services</h2>
        <ul>
            <li>Service 1</li>
            <li>Service 2</li>
            <li>Service 3</li>
        </ul>
    </section>

    <section>
        <h2>Contact Us</h2>
        <p>You can reach us at \
<a href="mailto:contact@example.com">contact@example.com</a>.</p>
    </section>

    <footer>
        <p>&copy; 2023 My Website</p>
    </footer>
</body>
</html>
"""


def test_html_tag_reader(html_str: str) -> None:
    with tempfile.NamedTemporaryFile(
        mode="w", delete=False, suffix=".html"
    ) as temp_file:
        temp_file.write(html_str)
        temp_file_path = Path(temp_file.name)

    reader = HTMLTagReader(ignore_no_id=True)
    docs = reader.load_data(temp_file_path)
    assert len(docs) == 2
    assert docs[0].metadata["tag_id"] == "about"
    assert docs[1].metadata["tag_id"] == "services"

    reader = HTMLTagReader()
    docs = reader.load_data(temp_file_path)
    assert len(docs) == 3
    assert docs[2].metadata["tag_id"] is None

    os.remove(temp_file.name)
